import os.path
import platform
import textwrap

import pytest

from conan.test.utils.mocks import ConanFileMock
from conan.tools.env.environment import environment_wrap_command
from conan.test.utils.tools import TestClient
from conan.internal.util.files import rmdir


@pytest.fixture(scope="module")
def transitive_shared_client():
    # TODO: Reuse fixtures
    client = TestClient(default_server_user=True)
    client.run("new cmake_lib -d name=hello -d version=0.1")
    client.run("create . -o hello/*:shared=True -tf=")
    client.save({}, clean_first=True)
    client.run("new cmake_lib -d name=chat -d version=0.1 -d requires=hello/0.1")
    client.run("create . -o chat/*:shared=True -o hello/*:shared=True -tf=")

    client.save({}, clean_first=True)
    client.run("new cmake_exe -d name=app -d version=0.1 -d requires=chat/0.1")
    client.run("create . -o chat/*:shared=True -o hello/*:shared=True -tf=")
    client.run("upload * -c -r default")
    client.run("remove * -c")
    return client


@pytest.mark.tool("cmake")
def test_other_client_can_execute(transitive_shared_client):
    _check_install_run(transitive_shared_client)


def _check_install_run(client):
    client = TestClient(servers=client.servers)
    client.run("install --requires=app/0.1@ -o chat*:shared=True -o hello/*:shared=True "
               "-g VirtualRunEnv")
    # This only finds "app" executable because the "app/0.1" is declaring package_type="application"
    # otherwise, run=None and nothing can tell us if the conanrunenv should have the PATH.
    command = environment_wrap_command(ConanFileMock(), "conanrun", client.current_folder, "app")

    client.run_command(command)
    assert "app/0.1: Hello World Release!" in client.out
    assert "chat/0.1: Hello World Release!" in client.out
    assert "hello/0.1: Hello World Release!" in client.out


@pytest.mark.tool("cmake")
def test_other_client_can_link_cmake(transitive_shared_client):
    client = transitive_shared_client
    # https://github.com/conan-io/conan/issues/13000
    # This failed, because of rpath link in Linux
    client = TestClient(servers=client.servers, inputs=["admin", "password"])
    client.run("new cmake_exe -d name=app -d version=0.1 -d requires=chat/0.1")
    client.run("create . -o chat/*:shared=True -o hello/*:shared=True -tf=")

    # check exe also keep running
    client.run("upload * -c -r default")
    client.run("remove * -c")
    _check_install_run(transitive_shared_client)


# FIXME: Move to the correct Meson space
@pytest.mark.tool("meson")
@pytest.mark.tool("ninja")
@pytest.mark.tool("pkg_config")
def test_other_client_can_link_meson(transitive_shared_client):
    client = transitive_shared_client
    # https://github.com/conan-io/conan/issues/13000
    # This failed, because of rpath link in Linux
    client = TestClient(servers=client.servers, inputs=["admin", "password"], path_with_spaces=False)
    client.run("new meson_exe -d name=app -d version=0.1 -d requires=chat/0.1")
    client.run("create . -o chat/*:shared=True -o hello/*:shared=True")
    # TODO Check that static builds too
    # client.run("create . --build=missing")


# FIXME: Move to the correct Meson space
@pytest.mark.tool("autotools")
@pytest.mark.skipif(platform.system() == "Windows", reason="Autotools needed")
def test_other_client_can_link_autotools(transitive_shared_client):
    client = transitive_shared_client
    # https://github.com/conan-io/conan/issues/13000
    # This failed, because of rpath link in Linux
    client = TestClient(servers=client.servers, inputs=["admin", "password"], path_with_spaces=False)
    client.run("new autotools_exe -d name=app -d version=0.1 -d requires=chat/0.1")
    client.run("create . -o chat/*:shared=True -o hello/*:shared=True")
    # TODO Check that static builds too
    # client.run("create . --build=missing")


@pytest.mark.tool("cmake")
def test_shared_cmake_toolchain_components():
    """ the same as above, but with components.
    """
    client = TestClient(default_server_user=True)

    client.run("new cmake_lib -d name=hello -d version=0.1")
    conanfile = client.load("conanfile.py")
    conanfile2 = conanfile.replace('self.cpp_info.libs = ["hello"]',
                                   'self.cpp_info.components["hi"].libs = ["hello"]')
    assert conanfile != conanfile2
    client.save({"conanfile.py": conanfile2})
    client.run("create . -o hello/*:shared=True -tf=")
    # Chat
    client.save({}, clean_first=True)
    client.run("new cmake_lib -d name=chat -d version=0.1 -d requires=hello/0.1")
    conanfile = client.load("conanfile.py")
    conanfile2 = conanfile.replace('self.cpp_info.libs = ["chat"]',
                                   'self.cpp_info.components["talk"].libs = ["chat"]\n'
                                   '        self.cpp_info.components["talk"].requires=["hello::hi"]')
    assert conanfile != conanfile2
    client.save({"conanfile.py": conanfile2})
    client.run("create . -o chat/*:shared=True -o hello/*:shared=True -tf=")

    # App
    client.save({}, clean_first=True)
    client.run("new cmake_exe -d name=app -d version=0.1 -d requires=chat/0.1")
    cmakelist = client.load("CMakeLists.txt")
    cmakelist2 = cmakelist.replace('target_link_libraries(app PRIVATE chat::chat)',
                                   'target_link_libraries(app PRIVATE chat::talk)')
    assert cmakelist != cmakelist2
    client.save({"CMakeLists.txt": cmakelist2})
    client.run("create . -o chat/*:shared=True -o hello/*:shared=True -tf=")
    client.run("upload * -c -r default")
    client.run("remove * -c")

    client = TestClient(servers=client.servers)
    client.run("install --requires=app/0.1@ -o chat*:shared=True -o hello/*:shared=True")
    # This only finds "app" executable because the "app/0.1" is declaring package_type="application"
    # otherwise, run=None and nothing can tell us if the conanrunenv should have the PATH.
    command = environment_wrap_command(ConanFileMock(), "conanrun", client.current_folder, "app")

    client.run_command(command)
    assert "app/0.1 test_package" in client.out
    assert "app/0.1: Hello World Release!" in client.out
    assert "chat/0.1: Hello World Release!" in client.out
    assert "hello/0.1: Hello World Release!" in client.out

    # https://github.com/conan-io/conan/issues/13000
    # This failed, because of rpath link in Linux
    client = TestClient(servers=client.servers, inputs=["admin", "password"])
    client.run("new cmake_exe -d name=app -d version=0.1 -d requires=chat/0.1")
    client.run("create . -o chat/*:shared=True -o hello/*:shared=True")
    client.run("upload * -c -r default")
    client.run("remove * -c")

    client = TestClient(servers=client.servers)
    client.run("install --requires=app/0.1@ -o chat*:shared=True -o hello/*:shared=True")
    # This only finds "app" executable because the "app/0.1" is declaring package_type="application"
    # otherwise, run=None and nothing can tell us if the conanrunenv should have the PATH.
    command = environment_wrap_command(ConanFileMock(), "conanrun", client.current_folder, "app")

    client.run_command(command)
    assert "app/0.1: Hello World Release!" in client.out
    assert "chat/0.1: Hello World Release!" in client.out
    assert "hello/0.1: Hello World Release!" in client.out


@pytest.mark.tool("cmake")
def test_shared_cmake_toolchain_test_package():
    # TODO: This is already tested in other places
    client = TestClient()
    client.run("new cmake_lib -d name=hello -d version=0.1")
    client.run("create . -o hello/*:shared=True")
    assert "hello/0.1: Hello World Release!" in client.out
    assert "hello/0.1 test_package" in client.out


@pytest.fixture()
def test_client_shared():
    client = TestClient()
    client.run("new -d name=hello -d version=0.1 cmake_lib")
    test_conanfile = textwrap.dedent("""
                import os
                from conan import ConanFile
                from conan.tools.cmake import CMake, cmake_layout
                from conan.tools.files import copy

                class Pkg(ConanFile):
                    settings = "os", "compiler", "arch", "build_type"
                    generators = "CMakeToolchain", "CMakeDeps"

                    def requirements(self):
                        self.requires(self.tested_reference_str)

                    def layout(self):
                        cmake_layout(self)

                    def build(self):
                        cmake = CMake(self)
                        cmake.configure()
                        cmake.build()

                    def generate(self):
                        for dep in self.dependencies.values():
                            copy(self, "*.dylib", dep.cpp_info.libdirs[0], self.build_folder)
                            copy(self, "*.dll", dep.cpp_info.libdirs[0], self.build_folder)

                    def test(self):
                        cmd = os.path.join(self.cpp.build.bindirs[0], "example")
                        # This is working without runenv because CMake is puting an internal rpath
                        # to the executable pointing to the dylib of hello, internally is doing something
                        # like: install_name_tool -add_rpath /path/to/hello/lib/libhello.dylib test
                        self.run(cmd)
                """)
    files = {"test_package/conanfile.py": test_conanfile}

    client.save(files)
    client.run("create . -o hello*:shared=True")
    assert "Hello World Release!" in client.out

    # We can run the exe from the test package directory also, without environment
    # because there is an internal RPATH in the exe with an abs path to the "hello"
    build_folder = client.created_test_build_folder("hello/0.1")
    exe_folder = os.path.join("test_package", build_folder)
    client.test_exe_folder = exe_folder
    assert os.path.exists(os.path.join(client.current_folder, exe_folder, "example"))
    client.run_command(os.path.join(exe_folder, "example"))

    # We try to remove the hello package and run again the executable from the test package,
    # this time it should fail, it doesn't find the shared library
    client.run("remove '*' -c")
    client.run_command(os.path.join(exe_folder, "example"), assert_error=True)
    return client


@pytest.mark.tool("cmake")
@pytest.mark.skipif(platform.system() != "Darwin", reason="Only OSX")
def test_shared_same_dir_using_tool(test_client_shared):
    """
    If we build an executable in Mac and we want it to locate the shared libraries in the same
    directory, we have different alternatives, here we use the "install_name_tool"
    """
    exe_folder = test_client_shared.test_exe_folder
    # Alternative 1, add the "." to the rpaths so the @rpath from the exe can be replaced with "."
    test_client_shared.current_folder = os.path.join(test_client_shared.current_folder, exe_folder)
    test_client_shared.run_command("install_name_tool -add_rpath '.' example")
    test_client_shared.run_command("./{}".format("example"))


@pytest.mark.tool("cmake")
@pytest.mark.skipif(platform.system() != "Darwin", reason="Only OSX")
def test_shared_same_dir_using_cmake(test_client_shared):
    """
        If we build an executable in Mac and we want it to locate the shared libraries in the same
        directory, we have different alternatives, here we use CMake to adjust CMAKE_INSTALL_RPATH
        to @executable_path so the exe knows that can replace @rpath with the current dir
    """

    # Alternative 2, set the rpath in cmake
    # Only viable when installing with cmake
    cmake = """
    set(CMAKE_CXX_COMPILER_WORKS 1)
    set(CMAKE_CXX_ABI_COMPILED 1)
    set(CMAKE_C_COMPILER_WORKS 1)
    set(CMAKE_C_ABI_COMPILED 1)
    cmake_minimum_required(VERSION 3.15)
    project(project CXX)

    set(CMAKE_INSTALL_RPATH "@executable_path")

    find_package(hello)
    add_executable(test  src/example.cpp )
    target_link_libraries(test  hello::hello)
    # Hardcoded installation path to keep the exe in the same place in the tests
    install(TARGETS test DESTINATION "bin")
    """
    # Same test conanfile but calling cmake.install()
    cf = textwrap.dedent("""
                import os
                from conan import ConanFile
                from conan.tools.files import copy
                from conan.tools.cmake import CMake, cmake_layout

                class Pkg(ConanFile):
                    settings = "os", "compiler", "arch", "build_type"
                    generators = "CMakeToolchain", "CMakeDeps"

                    def generate(self):
                        # The exe is installed by cmake at test_package/bin
                        dest = os.path.join(self.recipe_folder, "bin")
                        for dep in self.dependencies.values():
                            copy(self, "*.dylib", dep.cpp_info.libdirs[0], dest)

                    def requirements(self):
                        self.requires(self.tested_reference_str)

                    def layout(self):
                        cmake_layout(self)

                    def build(self):
                        cmake = CMake(self)
                        cmake.configure()
                        cmake.build()
                        cmake.install()

                    def test(self):
                        cmd = os.path.join(self.cpp.build.bindirs[0], "test")
                        # This is working without runenv because CMake is puting an internal rpath
                        # to the executable pointing to the dylib of hello, internally is doing something
                        # like: install_name_tool -add_rpath /path/to/hello/lib/libhello.dylib test
                        self.run(cmd)
                """)
    test_client_shared.save({"test_package/CMakeLists.txt": cmake, "test_package/conanfile.py": cf})
    test_client_shared.run("create . -o hello*:shared=True")
    test_client_shared.run("remove '*' -c")
    exe_folder = os.path.join("test_package", "bin")
    test_client_shared.run_command(os.path.join(exe_folder, "test"))


@pytest.mark.tool("cmake")
@pytest.mark.skipif(platform.system() != "Darwin", reason="Only OSX")
def test_shared_same_dir_using_env_var_current_dir(test_client_shared):
    """
        If we build an executable in Mac and we want it to locate the shared libraries in the same
        directory, we have different alternatives, here we set DYLD_LIBRARY_PATH before calling
        the executable but running in current dir
    """

    # Alternative 3, FAILING IN CI, set DYLD_LIBRARY_PATH in the current dir
    exe_folder = test_client_shared.test_exe_folder
    rmdir(os.path.join(test_client_shared.current_folder, exe_folder))
    test_client_shared.run("create . -o hello*:shared=True")
    test_client_shared.run("remove '*' -c")
    test_client_shared.current_folder = os.path.join(test_client_shared.current_folder, exe_folder)
    test_client_shared.run_command("DYLD_LIBRARY_PATH=$(pwd) ./example")
    test_client_shared.run_command("DYLD_LIBRARY_PATH=. ./example")
    # This assert is not working in CI, only locally
    # test_client_shared.run_command("DYLD_LIBRARY_PATH=@executable_path ./test")



from conan.test.utils.tools import GenConanfile, TestClient


class TestInstallParallel:

    def test_basic_parallel_install(self):
        client = TestClient(default_server_user=True)
        threads = 4
        counter = 8

        client.save_home({"global.conf": f"core.download:parallel={threads}"})
        client.save({"conanfile.py": GenConanfile()})

        for i in range(counter):
            client.run("create . --name=pkg%s --version=0.1 --user=user --channel=testing" % i)
        client.run("upload * --confirm -r default")
        client.run("remove * -c")

        # Lets consume the packages
        conanfile_txt = ["[requires]"]
        for i in range(counter):
            conanfile_txt.append("pkg%s/0.1@user/testing" % i)
        conanfile_txt = "\n".join(conanfile_txt)

        client.save({"conanfile.txt": conanfile_txt}, clean_first=True)
        client.run("install .")
        assert "Downloading binary packages in %s parallel threads" % threads in client.out
        for i in range(counter):
            assert "pkg%s/0.1@user/testing: Package installed" % i in client.out



import os
import textwrap

from conan.internal.model.info import load_binary_info
from conan.api.model import RecipeReference
from conan.internal.paths import CONANFILE
from conan.test.utils.tools import TestClient
from conan.internal.util.files import load, save


class TestSettings:

    def _get_conaninfo(self, reference, client):
        ref = client.cache.get_latest_recipe_revision(RecipeReference.loads(reference))
        pkg_ids = client.cache.get_package_references(ref)
        pref = client.cache.get_latest_package_revision(pkg_ids[0])
        pkg_folder = client.cache.pkg_layout(pref).package()
        return load_binary_info(client.load(os.path.join(pkg_folder, "conaninfo.txt")))

    def test_wrong_settings(self):
        settings = """os:
    null:
        subsystem: [null, msys]
"""
        client = TestClient()
        save(client.paths.settings_path, settings)
        client.save_home({"profiles/default": ""})
        conanfile = """from conan import ConanFile
class Pkg(ConanFile):
    settings = "os", "compiler"
"""
        client.save({"conanfile.py": conanfile})
        client.run("create . --name=pkg --version=0.1 --user=lasote --channel=testing", assert_error=True)
        assert "ERROR: settings.yml: null setting can't have subsettings" in client.out

    def test_settings_constraint_error_type(self):
        # https://github.com/conan-io/conan/issues/3022
        conanfile = """from conan import ConanFile
class Test(ConanFile):
    settings = "os"
    def build(self):
        self.output.info("OS!!: %s" % self.settings.os)
    """
        client = TestClient()
        client.save({"conanfile.py": conanfile})
        client.run("create . --name=pkg --version=0.1 --user=user --channel=testing -s os=Linux")
        assert "pkg/0.1@user/testing: OS!!: Linux" in client.out

    def test_settings_as_a_str(self):
        content = """
from conan import ConanFile

class SayConan(ConanFile):
    name = "say"
    version = "0.1"
    settings = "os"
"""
        client = TestClient()
        client.save({CONANFILE: content})
        client.run("create . -s os=Windows --build missing")
        # Now read the conaninfo and verify that settings applied is only os and value is windows
        conan_info = self._get_conaninfo("say/0.1@", client)
        assert conan_info["settings"]["os"] == "Windows"

        client.run("remove say/0.1 -c")
        client.run("create . -s os=Linux --build missing")
        # Now read the conaninfo and verify that settings applied is only os and value is windows
        conan_info = self._get_conaninfo("say/0.1@", client)
        assert conan_info["settings"]["os"] == "Linux"

    def test_settings_as_a_list_conanfile(self):
        # Now with conanfile as a list
        content = """
from conan import ConanFile

class SayConan(ConanFile):
    name = "say"
    version = "0.1"
    settings = "os", "arch"
"""
        client = TestClient()
        client.save({CONANFILE: content})
        client.run("create . -s os=Windows --build missing")
        conan_info = self._get_conaninfo("say/0.1@", client)
        assert conan_info["settings"]["os"] == "Windows"

    def test_settings_as_a_dict_conanfile(self):
        # Now with conanfile as a dict
        # XXX: this test only works on machines w default arch "x86", "x86_64", "sparc" or "sparcv9"
        content = """
from conan import ConanFile

class SayConan(ConanFile):
    name = "say"
    version = "0.1"
    settings = {"os", "arch"}
"""
        client = TestClient()
        client.save({CONANFILE: content})
        client.run("create . -s os=Windows --build missing")
        conan_info = self._get_conaninfo("say/0.1@", client)
        assert conan_info["settings"]["os"] == "Windows"

    def test_invalid_settings3(self):
        client = TestClient()

        # Test wrong settings in conanfile
        content = textwrap.dedent("""
            from conan import ConanFile

            class SayConan(ConanFile):
                settings = "invalid"
            """)

        client.save({CONANFILE: content})
        client.run("install . --build missing", assert_error=True)
        assert "'settings.invalid' doesn't exist" in client.out

        # Test wrong values in conanfile
    def test_invalid_settings4(self):
        content = """
from conan import ConanFile

class SayConan(ConanFile):
    name = "say"
    version = "0.1"
    settings = "os"
"""
        client = TestClient()
        client.save({CONANFILE: content})
        client.run("create . -s os=ChromeOS --build missing", assert_error=True)
        assert "ERROR: Invalid setting 'ChromeOS' is not a valid 'settings.os' value." in client.out
        assert "Possible values are ['Windows', 'WindowsStore', 'WindowsCE', 'Linux'" in client.out

        # Now add new settings to config and try again
        config = load(client.paths.settings_path)
        config = config.replace("Windows:",
                                "Windows:\n    ChromeOS:\n")

        save(client.paths.settings_path, config)
        client.run("create . -s os=ChromeOS --build missing")

        # Settings is None
        content = """
from conan import ConanFile

class SayConan(ConanFile):
    name = "say"
    version = "0.1"
    settings = None
"""
        client.save({CONANFILE: content})
        client.run("remove say/0.1 -c")
        client.run("create . --build missing")
        conan_info = self._get_conaninfo("say/0.1", client)
        assert conan_info.get("settings") == None

        # Settings is {}
        content = """
from conan import ConanFile

class SayConan(ConanFile):
    name = "say"
    version = "0.1"
    settings = {}
"""
        client.save({CONANFILE: content})
        client.run("remove say/0.1 -c")
        client.run("create . --build missing")
        conan_info = self._get_conaninfo("say/0.1", client)

        assert conan_info.get("settings") == None



import pytest

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


@pytest.mark.parametrize("compiler, version, cppstd, correct", [
    ("gcc", "4.1", "11", False),
    ("gcc", "4.4", "11", True),
    ("gcc", "4.4", "14", False),
    ("gcc", "4.4", "17", False),
    ("gcc", "4.4", "20", False),
    ("gcc", "4.8", "11", True),
    ("gcc", "4.8", "14", True),
    ("gcc", "4.8", "17", False),
    ("gcc", "4.8", "20", False),
    ("gcc", "5", "17", True),
    ("gcc", "5", "20", False),
    ("gcc", "8", "17", True),
    ("gcc", "8", "20", True),
    ("clang", "3.3", "11", True),
    ("clang", "3.3", "14", False),
    ("clang", "3.4", "14", True),
    ("clang", "3.4", "17", False),
    ("clang", "3.5", "17", True),
    ("clang", "3.5", "20", False),
    ("clang", "5.0", "20", False),
    ("clang", "6.0", "20", True),
    ("apple-clang", "5.0", "98", True),
    ("apple-clang", "5.0", "11", True),
    ("apple-clang", "5.1", "14", True),
    ("apple-clang", "5.1", "17", False),
    ("apple-clang", "6.1", "17", True),
    ("apple-clang", "6.1", "20", False),
    ("apple-clang", "10.0", "20", True),
    ("msvc", "170", "14", False),
    ("msvc", "190", "14", True),
    ("msvc", "190", "20", False),
    ("msvc", "191", "14", True),
    ("msvc", "191", "17", True),
    ("msvc", "191", "20", False),
    ("msvc", "193", "20", True)
])
def test_invalid_cppstd(compiler, version, cppstd, correct):
    c = TestClient()
    c.save({"conanfile.py": GenConanfile()})
    c.run("install . -s compiler={} -s compiler.version={} "
          "-s compiler.cppstd={}".format(compiler, version, cppstd), assert_error=not correct)
    if not correct:
        assert "ConanException: The provided compiler.cppstd" in c.out



import os
import platform
import shutil

import pytest

from conan.tools.env.environment import environment_wrap_command
from conan.tools.files import replace_in_file
from conan.test.utils.mocks import ConanFileMock
from conan.test.utils.tools import TestClient


@pytest.mark.skipif(platform.system() != "Linux", reason="Only Linux")
@pytest.mark.tool("cmake")
@pytest.mark.parametrize("nosoname_property", [
    True,  # without SONAME
    False  # By default, with SONAME
])
def test_no_soname_flag(nosoname_property):
    """ This test case is testing this graph structure:
            *   'Executable' -> 'LibB' -> 'LibNoSoname'
        Where:
            *   LibNoSoname: is a package built as shared and without the SONAME flag.
            *   LibB: is a package which requires LibNoSoname.
            *   Executable: is the final consumer building an application and depending on OtherLib.
        How:
            1- Creates LibNoSoname and upload it to remote server
            2- Creates LibB and upload it to remote server
            3- Remove the Conan cache folder
            4- Creates an application and consume LibB
        Goal:
            * If `self.cpp_info.set_property("nosoname", True), then the `Executable` runs OK.
            * If `self.cpp_info.set_property("nosoname", False), then the `Executable` fails.
    """
    client = TestClient(default_server_user=True)
    # Creating nosoname/0.1 library
    client.run("new cmake_lib -d name=nosoname -d version=0.1")

    replace_in_file(ConanFileMock(), os.path.join(client.current_folder, "conanfile.py"),
                    'self.cpp_info.libs = ["nosoname"]',
                    f'self.cpp_info.libs = ["nosoname"]\n        self.cpp_info.set_property("nosoname", {nosoname_property})')
    replace_in_file(ConanFileMock(), os.path.join(client.current_folder, "CMakeLists.txt"),
                    'target_include_directories(nosoname PUBLIC include)',
                    'target_include_directories(nosoname PUBLIC include)\nset_target_properties(nosoname PROPERTIES NO_SONAME 1)')
    client.run("create . -o nosoname/*:shared=True -tf=")
    # Creating lib_b/0.1 library (depends on nosoname/0.1)
    client.save({}, clean_first=True)
    client.run("new cmake_lib -d name=lib_b -d version=0.1 -d requires=nosoname/0.1")
    client.run("create . -o lib_b/*:shared=True -o nosoname/*:shared=True -tf=")
    # Creating app/0.1 application (depends on lib_b/0.1)
    client.save({}, clean_first=True)
    client.run("new cmake_exe -d name=app -d version=0.1 -d requires=lib_b/0.1")
    client.run("create . -o nosoname/*:shared=True -o lib_b/*:shared=True -tf=")
    client.run("upload * -c -r default")
    # Removing everything from the .conan2/p to ensure that we don't have anything saved in the cache
    shutil.rmtree(client.cache.store)

    client = TestClient(servers=client.servers)
    client.run("install --requires=app/0.1@ -o nosoname*:shared=True -o lib_b/*:shared=True")
    # This only finds "app" executable because the "app/0.1" is declaring package_type="application"
    # otherwise, run=None and nothing can tell us if the conanrunenv should have the PATH.
    command = environment_wrap_command(ConanFileMock(), "conanrun", client.current_folder, "app")

    # If `nosoname_property` is False, and we have a library without the SONAME flag,
    # then it should fail
    if nosoname_property is False:
        client.run_command(command, assert_error=True)
        assert "libnosoname.so: cannot open shared object file: " \
               "No such file or directory" in client.out
    else:
        client.run_command(command)
        assert "nosoname/0.1: Hello World Release!" in client.out
        assert "lib_b/0.1: Hello World Release!" in client.out
        assert "app/0.1: Hello World Release!" in client.out



import os
import textwrap
import platform

import pytest

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


def test_rosenv():
    """
    Test that the ROSEnv generator generates the environment files and that
    the environment variables are correctly set.
    """
    client = TestClient()
    conanfile3 = textwrap.dedent('''
        [requires]
        [generators]
        CMakeDeps
        CMakeToolchain
        ROSEnv
        ''')
    client.save({
        "conanfile3.txt": conanfile3
    })

    if platform.system() == "Windows":
        conanrosenv_file = "conanrosenv.bat"
        conanrosenv_build = "conanrosenv-build.bat"
        build_type = "Release"
        cmd_source = "call"
        cmd_env = "set"
    else:
        conanrosenv_file = "conanrosenv.sh"
        conanrosenv_build = "conanrosenv-build.sh"
        build_type = '"Release"'
        cmd_source = "."
        cmd_env = "env"


    client.run("install conanfile3.txt --output-folder install/conan")
    assert f"Generated ROSEnv Conan file: {conanrosenv_file}" in client.out
    install_folder = os.path.join(client.current_folder, "install", "conan")
    conanrosenv_path = os.path.join(install_folder, conanrosenv_file)
    assert os.path.exists(conanrosenv_path)
    conanrosenvbuild_path = os.path.join(install_folder, conanrosenv_build)
    assert os.path.exists(conanrosenvbuild_path)
    toolchain_path = os.path.join(client.current_folder, "install", "conan",
                                  "conan_toolchain.cmake")
    content = client.load(conanrosenvbuild_path)
    if platform.system() != "Windows":
        assert f'CMAKE_TOOLCHAIN_FILE="{toolchain_path}"' in content
    else:
        assert f'CMAKE_TOOLCHAIN_FILE={toolchain_path}' in content
    assert f'CMAKE_BUILD_TYPE={build_type}' in content

    client.run_command(f'{cmd_source} "{conanrosenv_path}" && {cmd_env}')
    assert f"CMAKE_TOOLCHAIN_FILE={toolchain_path}" in client.out
    assert "CMAKE_BUILD_TYPE=Release" in client.out


@pytest.mark.skipif(platform.system() == "Windows", reason="Uses UNIX commands")
def test_rosenv_shared_libraries():
    """
    Test that the library paths env vars are set up correctly so that the executables built with
    colcon can found the shared libraries of conan packages
    """
    client = TestClient()
    c1 = GenConanfile("lib1", "1.0").with_shared_option(False).with_package_file("lib/lib1", "lib-content")
    c2 = GenConanfile("lib2", "1.0").with_shared_option(False).with_requirement("lib1/1.0").with_package_file("lib/lib2", "lib-content")
    c3 = textwrap.dedent('''
           [requires]
           lib2/1.0
           [generators]
           CMakeDeps
           CMakeToolchain
           ROSEnv
           ''')
    client.save({
        "conanfile1.py": c1,
        "conanfile2.py": c2,
        "conanfile3.txt": c3
    })

    client.run("create conanfile1.py -o *:shared=True")
    client.run("create conanfile2.py -o *:shared=True")
    client.run("install conanfile3.txt -o *:shared=True --output-folder install/conan")
    conanrosenv_path = os.path.join(client.current_folder, "install", "conan", "conanrosenv.sh")
    client.run_command(f". \"{conanrosenv_path}\" && env")
    environment_content = client.out
    client.run(
        "cache path lib1/1.0#58723f478a96866dcbd9456d8eefd7c4:1744785cb24e3bdca70e27041dc5abd20476f947")
    lib1_lib_path = os.path.join(client.out.strip(), "lib")
    assert lib1_lib_path in environment_content
    client.run(
        "cache path lib2/1.0#4b7a6063ba107d770458ce10385beb52:5c3c2e56259489f7ffbc8e494921eda4b747ef21")
    lib2_lib_path = os.path.join(client.out.strip(), "lib")
    assert lib2_lib_path in environment_content



from conan.test.assets.genconanfile import GenConanfile
from test.integration.graph.core.graph_manager_base import GraphManagerTest
from test.integration.graph.core.graph_manager_test import _check_transitive


class TestOptions(GraphManagerTest):

    def test_basic(self):
        # app -> libb0.1 (lib shared=True) -> liba0.1 (default static)
        # By default if packages do not specify anything link=True is propagated run=None (unknown)
        self.recipe_cache("liba/0.1", option_shared=False)
        self.recipe_conanfile("libb/0.1", GenConanfile().with_requires("liba/0.1").
                              with_default_option("liba/*:shared", True))
        consumer = self.recipe_consumer("app/0.1", ["libb/0.1"])

        deps_graph = self.build_consumer(consumer)

        assert 3 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libb])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[app])
        self._check_node(liba, "liba/0.1#123", dependents=[libb], options={"shared": "True"})

        # node, include, link, build, run
        _check_transitive(app, [(libb, True, True, False, False),
                                (liba, True, True, False, True)])
        _check_transitive(libb, [(liba, True, True, False, True)])

    def test_app_override(self):
        # app (liba static)-> libb0.1 (liba shared) -> liba0.1 (default static)
        # By default if packages do not specify anything link=True is propagated run=None (unknown)
        self.recipe_cache("liba/0.1", option_shared=False)
        self.recipe_conanfile("libb/0.1", GenConanfile().with_requires("liba/0.1").
                              with_default_option("liba/*:shared", True))
        consumer = self.consumer_conanfile(GenConanfile("app", "0.1").with_requires("libb/0.1").
                                           with_default_option("liba/*:shared", False))

        deps_graph = self.build_consumer(consumer)

        assert 3 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libb])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[app])
        self._check_node(liba, "liba/0.1#123", dependents=[libb], options={"shared": "False"})

        # node, include, link, build, run
        _check_transitive(app, [(libb, True, True, False, False),
                                (liba, True, True, False, False)])
        _check_transitive(libb, [(liba, True, True, False, False)])

    def test_diamond_no_conflict(self):
        # app -> libb0.1 ---------------> liba0.1
        #    \-> libc0.1 (liba shared) -> liba0.1
        # Just the first expanded one will prevail
        self.recipe_cache("liba/0.1", option_shared=False)
        self.recipe_cache("libb/0.1", ["liba/0.1"])
        self.recipe_conanfile("libc/0.1", GenConanfile().with_requires("liba/0.1").
                              with_default_option("liba*:shared", True))

        # If we expand first the "libb", it will expand to "liba" as static (shared=False)
        consumer = self.recipe_consumer("app/0.1", ["libb/0.1", "libc/0.1"])
        deps_graph = self.build_consumer(consumer)
        assert 4 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst
        libc = app.edges[1].dst
        liba = libb.edges[0].dst
        liba2 = libc.edges[0].dst
        assert liba is liba2
        self._check_node(app, "app/0.1", deps=[libc, libb])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[app])
        self._check_node(liba, "liba/0.1#123", dependents=[libb, libc], options={"shared": "False"})

        # Changing the order, will make "liba" shared=True
        consumer = self.recipe_consumer("app/0.1", ["libc/0.1", "libb/0.1"])
        deps_graph = self.build_consumer(consumer)
        assert 4 == len(deps_graph.nodes)
        app = deps_graph.root
        libc = app.edges[0].dst
        libb = app.edges[1].dst
        liba = libb.edges[0].dst
        liba2 = libc.edges[0].dst
        assert liba is liba2
        self._check_node(app, "app/0.1", deps=[libc, libb])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[app])
        self._check_node(liba, "liba/0.1#123", dependents=[libb, libc], options={"shared": "True"})

    def test_diamond_downstream_priority(self):
        # app(liba:shared) -> libb0.1 ---------------> liba0.1
        #    \--------------> libc0.1 (liba shared) -> liba0.1
        self.recipe_cache("liba/0.1", option_shared=False)
        self.recipe_cache("libb/0.1", ["liba/0.1"])
        self.recipe_conanfile("libc/0.1", GenConanfile().with_requires("liba/0.1").
                              with_default_option("liba*:shared", True))

        for shared in True, False:
            consumer = self.consumer_conanfile(GenConanfile("app", "0.1")
                                               .with_requires("libb/0.1", "libc/0.1")
                                               .with_default_option("liba*:shared", shared))
            deps_graph = self.build_consumer(consumer)

            assert 4 == len(deps_graph.nodes)
            app = deps_graph.root
            libb = app.edges[0].dst
            libc = app.edges[1].dst
            liba = libb.edges[0].dst
            liba2 = libc.edges[0].dst
            assert liba is liba2
            self._check_node(app, "app/0.1", deps=[libc, libb])
            self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[app])
            self._check_node(liba, "liba/0.1#123", dependents=[libb, libc],
                             options={"shared": str(shared)})

            # order doesn't matter
            consumer = self.consumer_conanfile(GenConanfile("app", "0.1")
                                               .with_requires("libc/0.1", "libb/0.1")
                                               .with_default_option("liba*:shared", shared))
            deps_graph = self.build_consumer(consumer)

            assert 4 == len(deps_graph.nodes)
            app = deps_graph.root
            libc = app.edges[0].dst
            libb = app.edges[1].dst
            liba = libb.edges[0].dst
            liba2 = libc.edges[0].dst
            assert liba is liba2
            self._check_node(app, "app/0.1", deps=[libc, libb])
            self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[app])
            self._check_node(liba, "liba/0.1#123", dependents=[libb, libc],
                             options={"shared": str(shared)})


class TestBuildRequireOptions(GraphManagerTest):
    def test_protobuf_different_options_profile(self):
        # app -> lib ------> protobuf -> zlib (shared)
        #          \--(br)-> protobuf -> zlib (static)
        self._cache_recipe("zlib/0.1", GenConanfile().with_shared_option(True))
        self._cache_recipe("protobuf/0.1", GenConanfile().with_require("zlib/0.1"))
        self._cache_recipe("lib/0.1", GenConanfile().with_requires("protobuf/0.1").
                           with_tool_requires("protobuf/0.1"))
        deps_graph = self.build_graph(GenConanfile("app", "0.1").with_require("lib/0.1"),
                                      options_build={"zlib*:shared": False})

        assert 6 == len(deps_graph.nodes)
        app = deps_graph.root
        lib = app.edges[0].dst
        protobuf_host = lib.edges[0].dst
        protobuf_build = lib.edges[1].dst
        zlib_shared = protobuf_host.edges[0].dst
        zlib_static = protobuf_build.edges[0].dst

        self._check_node(app, "app/0.1@", deps=[lib], dependents=[])
        self._check_node(lib, "lib/0.1#123", deps=[protobuf_host, protobuf_build], dependents=[app])
        self._check_node(protobuf_host, "protobuf/0.1#123", deps=[zlib_shared], dependents=[lib])
        self._check_node(protobuf_build, "protobuf/0.1#123", deps=[zlib_static], dependents=[lib])
        self._check_node(zlib_shared, "zlib/0.1#123", deps=[], dependents=[protobuf_host])
        self._check_node(zlib_static, "zlib/0.1#123", deps=[], dependents=[protobuf_build])
        assert not zlib_static.conanfile.options.shared
        assert zlib_shared.conanfile.options.shared

        # node, include, link, build, run
        _check_transitive(app, [(lib, True, True, False, False),
                                (protobuf_host, True, True, False, False),
                                (zlib_shared, True, True, False, True)])
        _check_transitive(lib, [(protobuf_host, True, True, False, False),
                                (zlib_shared, True, True, False, True),
                                (protobuf_build, False, False, True, True)])



import copy
import os
import pytest

from conan.internal.errors import NotFoundException
from conan.internal.model.manifest import FileTreeManifest
from conan.api.model import PkgReference
from conan.api.model import RecipeReference
from conan.internal.paths import CONANINFO, CONAN_MANIFEST
from conans.server.service.authorize import BasicAuthorizer
from conans.server.service.v2.search import SearchService
from conans.server.service.v2.service_v2 import ConanServiceV2
from conans.server.store.disk_adapter import ServerDiskAdapter
from conans.server.store.server_store import ServerStore
from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.test_files import temp_folder
from conan.internal.util.files import save, save_files


DEFAULT_REVISION = "1234"


class TestConanService:

    @pytest.fixture(autouse=True)
    def setup(self):
        self.ref = RecipeReference.loads("openssl/2.0.3@lasote/testing#%s" % DEFAULT_REVISION)

        self.pref = PkgReference(self.ref, "123123123", DEFAULT_REVISION)
        self.tmp_dir = temp_folder()

        read_perms = [("*/*@*/*", "*")]
        write_perms = [("*/*@*/*", "*")]
        authorizer = BasicAuthorizer(read_perms, write_perms)

        self.fake_url = "http://url"
        adapter = ServerDiskAdapter(self.fake_url, self.tmp_dir)
        self.server_store = ServerStore(storage_adapter=adapter)
        self.service = ConanServiceV2(authorizer, self.server_store)
        self.search_service = SearchService(authorizer, self.server_store, "lasote")

        files = {"conanfile.py": str(GenConanfile("test"))}
        save_files(self.server_store.export(self.ref), files)
        self.server_store.update_last_revision(self.ref)
        manifest = FileTreeManifest.create(self.server_store.export(self.ref))
        conan_digest_path = os.path.join(self.server_store.export(self.ref), CONAN_MANIFEST)
        save(conan_digest_path, repr(manifest))

        files = {"boost.lib": "", "boost2.lib": ""}
        save_files(self.server_store.package(self.pref), files)

    def test_search(self):
        """ check the dict is returned by get_packages_info service
        """
        # Creating and saving conans, packages, and conans.vars
        ref2 = RecipeReference("openssl", "3.0", "lasote", "stable", DEFAULT_REVISION)
        ref3 = RecipeReference("Assimp", "1.10", "fenix", "stable", DEFAULT_REVISION)
        ref4 = RecipeReference("assimpFake", "0.1", "phil", "stable", DEFAULT_REVISION)

        pref2 = PkgReference(ref2, "12345587754", DEFAULT_REVISION)
        pref3 = PkgReference(ref3, "77777777777", DEFAULT_REVISION)

        conan_vars = """
[options]
    use_Qt=%s
"""
        conan_vars1 = conan_vars % "True"
        conan_vars2 = conan_vars % "False"
        conan_vars3 = conan_vars % "True"

        save_files(self.server_store.package(self.pref), {CONANINFO: conan_vars1})
        self.server_store.update_last_package_revision(self.pref)
        save_files(self.server_store.package(pref2), {CONANINFO: conan_vars2})
        self.server_store.update_last_package_revision(pref2)
        save_files(self.server_store.package(pref3), {CONANINFO: conan_vars3})
        self.server_store.update_last_package_revision(pref3)

        save_files(self.server_store.export(ref4), {"dummy.txt": "//"})

        info = self.search_service.search()
        expected = [RecipeReference(r.name, r.version, r.user, r.channel, revision=None)
                    for r in [ref3, ref4, self.ref, ref2]]
        assert expected == info

        info = self.search_service.search(pattern="Assimp*", ignorecase=False)
        ref3_norev = copy.copy(ref3)
        ref3_norev.revision = None
        assert info == [ref3_norev]

        info = self.search_service.search_packages(ref2)
        assert info == {'12345587754': {'content': '\n[options]\n    use_Qt=False\n',
                                                }}

        info = self.search_service.search_packages(ref3)
        assert info == {'77777777777': {'content': '\n[options]\n    use_Qt=True\n'}
                                }

        info = self.search_service.search_packages(ref2, list_only=True)
        assert info == {'12345587754': {}}

        info = self.search_service.search_packages(ref3, list_only=True)
        assert info == {'77777777777': {}}

    def test_remove(self):
        ref2 = RecipeReference("OpenCV", "3.0", "lasote", "stable", DEFAULT_REVISION)
        ref3 = RecipeReference("Assimp", "1.10", "lasote", "stable", DEFAULT_REVISION)

        pref2 = PkgReference(ref2, "12345587754", DEFAULT_REVISION)
        pref3 = PkgReference(ref3, "77777777777", DEFAULT_REVISION)

        save_files(self.server_store.export(ref2), {"fake.txt": "//fake"})
        self.server_store.update_last_revision(ref2)
        save_files(self.server_store.package(pref2), {"fake.txt": "//fake"})
        self.server_store.update_last_package_revision(pref2)
        save_files(self.server_store.package(pref3), {"fake.txt": "//fake"})
        self.server_store.update_last_package_revision(pref3)

        # Delete all the conans folder
        self.service.remove_recipe(self.ref, "lasote")
        conan_path = self.server_store.base_folder(self.ref)
        assert not os.path.exists(conan_path)

        # Raise an exception
        with pytest.raises(NotFoundException):
            self.service.remove_recipe(
                              RecipeReference("Fake", "1.0", "lasote", "stable"), "lasote")



import os
import platform
import stat
import pytest

from conan.errors import ConanException
from conan.tools.files import chmod
from conan.test.utils.mocks import ConanFileMock
from conan.test.utils.tools import temp_folder, save_files


@pytest.mark.skipif(platform.system() == "Windows", reason="validate full permissions only in Unix")
@pytest.mark.parametrize("read,write,execute,expected", [
    (True, True, True, 0o700),
    (False, True, False, 0o200),
    (False, False, True, 0o100),
    (True, False, True, 0o500),
    (True, True, False, 0o600),
    (True, False, False, 0o400),
    (False, False, False, 0o000),])
def test_chmod_single_file(read, write, execute, expected):
    """
    The chmod should be able to change the permissions of a single file.
    """
    tmp = temp_folder()
    save_files(tmp, {"file.txt": "foobar"})
    file_path = os.path.join(tmp, "file.txt")
    os.chmod(file_path, 0o000)
    conanfile = ConanFileMock()
    chmod(conanfile, file_path, read=read, write=write, execute=execute, recursive=False)
    file_mode = os.stat(file_path).st_mode
    assert stat.S_IMODE(file_mode) == expected


@pytest.mark.skipif(platform.system() == "Windows", reason="validate full permissions only in Unix")
@pytest.mark.parametrize("read,write,execute,expected", [
    (True, True, True, 0o700),
    (False, True, False, 0o200),
    (False, False, True, 0o100),
    (True, False, True, 0o500),
    (True, True, False, 0o600),
    (True, False, False, 0o400),
    (False, False, False, 0o000),])
def test_chmod_recursive(read, write, execute, expected):
    """
    The chmod should be able to change the permissions of all files in a folder when recursive is set to True.
    """
    tmp = temp_folder()
    files = {"foobar/qux/file.txt": "foobar",
             "foobar/file.txt": "qux",
             "foobar/foo/file.txt": "foobar"}
    save_files(tmp, files)
    folder_path = os.path.join(tmp, "foobar")
    for file in files.keys():
        file_path = os.path.join(tmp, file)
        os.chmod(file_path, 0o000)
    conanfile = ConanFileMock()
    chmod(conanfile, folder_path, read=read, write=write, execute=execute, recursive=True)
    for file in files.keys():
        file_mode = os.stat(os.path.join(tmp, file)).st_mode
        assert stat.S_IMODE(file_mode) == expected


@pytest.mark.skipif(platform.system() == "Windows", reason="Validate default permissions only in Unix")
def test_chmod_default_values():
    """
    When not passing a permission parameter, chmod should not change the specific permission.
    """
    tmp = temp_folder()
    save_files(tmp, {"file.txt": "foobar"})
    file_path = os.path.join(tmp, "file.txt")
    os.chmod(file_path, 0o111)
    conanfile = ConanFileMock()
    chmod(conanfile, file_path, read=True)
    file_mode = os.stat(file_path).st_mode
    assert stat.S_IMODE(file_mode) == 0o511


def test_missing_permission_arguments():
    """
    The chmod should raise an exception if no new permission is provided.
    """
    conanfile = ConanFileMock()
    with pytest.raises(ConanException) as error:
        chmod(conanfile, "invalid_path")
    assert 'Could not change permission: At least one of the permissions should be set.' in str(error.value)


def test_invalid_path():
    """
    The chmod should raise an exception if the path does not exist.
    """
    conanfile = ConanFileMock()
    with pytest.raises(ConanException) as error:
        chmod(conanfile, "invalid_path", read=True, write=True, execute=True, recursive=False)
    assert 'Could not change permission: Path "invalid_path" does not exist.' in str(error.value)


@pytest.mark.skipif(platform.system() != "Windows", reason="Validate read-only permissions only in Windows")
def test_chmod_windows():
    """
    The chmod should be able to change read-only state in Windows.
    """
    tmp = temp_folder()
    save_files(tmp, {"file.txt": "foobar"})
    file_path = os.path.join(tmp, "file.txt")
    os.chmod(file_path, 0o000)
    conanfile = ConanFileMock()
    chmod(conanfile, file_path, read=True, write=True, execute=True, recursive=False)
    assert os.access(file_path, os.W_OK)



import platform
import re
import tempfile
import textwrap

import pytest

from conan.test.assets.cmake import gen_cmakelists
from conan.test.assets.sources import gen_function_cpp, gen_function_c
from test.functional.utils import check_vs_runtime, check_exe_run
from conan.test.utils.tools import TestClient


@pytest.fixture
def client():
    # IMPORTANT: This cannot use the default tests location, if in Windows, it can be another unit
    # like F and Visual WONT FIND ClangCL
    t = tempfile.mkdtemp(suffix='conans')
    c = TestClient(cache_folder=t)
    clang_profile = textwrap.dedent("""
        [settings]
        os=Windows
        arch=x86_64
        build_type=Release
        compiler=clang
        compiler.version=18
        compiler.cppstd=14
        [conf]
        tools.build:compiler_executables={"cpp": "clang++", "c": "clang", "rc": "clang"}
        """)
    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.cmake import CMake, cmake_layout

        class Pkg(ConanFile):
            settings = "os", "compiler", "build_type", "arch"
            exports_sources = "*"
            generators = "CMakeToolchain"

            def layout(self):
                cmake_layout(self)

            def build(self):
                cmake = CMake(self)
                cmake.configure()
                cmake.build()
                cmd = os.path.join(self.cpp.build.bindirs[0], "my_app")
                self.run(cmd)

            def package(self):
                cmake = CMake(self)
                cmake.install()

            def package_info(self):
                cmd = os.path.join(self.package_folder, "bin", "my_app")
                self.output.info("MYCMD={}!".format(os.path.abspath(cmd)))
        """)
    c.save({"conanfile.py": conanfile,
            "clang": clang_profile,
            "CMakeLists.txt": gen_cmakelists(appname="my_app", appsources=["src/main.cpp"],
                                             install=True),
            "src/main.cpp": gen_function_cpp(name="main")})
    return c


@pytest.mark.tool("cmake")
@pytest.mark.tool("clang", "18")
@pytest.mark.skipif(platform.system() != "Windows", reason="requires Win")
class TestLLVMClang:
    """ External LLVM/clang, with different CMake generators
    This links always with the VS runtime, it is built-in
    """

    @pytest.mark.tool("mingw64")
    @pytest.mark.tool("visual_studio", "17")
    @pytest.mark.tool("clang", "18")  # repeated, for priority over the mingw64 clang
    @pytest.mark.parametrize("runtime", ["static", "dynamic"])
    def test_clang_mingw(self, client, runtime):
        """ compiling with an LLVM-clang installed, which uses by default the
        VS runtime
        """
        client.run("create . --name=pkg --version=0.1 -pr=clang "
                   "-s compiler.runtime_version=v144 "
                   "-s compiler.runtime={}".format(runtime))
        # clang compilations in Windows will use MinGW Makefiles by default
        assert 'cmake -G "MinGW Makefiles"' in client.out
        assert "GNU-like command-line" in client.out
        assert "main __clang_major__18" in client.out
        assert "main _MSC_VER194" in client.out
        assert "main _MSVC_LANG2014" in client.out
        assert "main _M_X64 defined" in client.out
        assert "main __x86_64__ defined" in client.out

        check_exe_run(client.out, "main", "clang", None, "Release", "x86_64", "14")
        cmd = re.search(r"MYCMD=(.*)!", str(client.out)).group(1)
        cmd = cmd + ".exe"
        static_runtime = (runtime == "static")
        check_vs_runtime(cmd, client, "17", build_type="Release", static_runtime=static_runtime)

    @pytest.mark.tool("ninja")
    @pytest.mark.tool("visual_studio", "17")
    @pytest.mark.parametrize("generator", ["Ninja", "NMake Makefiles"])
    def test_clang_cmake_ninja_nmake(self, client, generator):
        client.run("create . --name=pkg --version=0.1 -pr=clang -s compiler.runtime=dynamic "
                   "-s compiler.runtime_version=v144 "
                   '-c tools.cmake.cmaketoolchain:generator="{}"'.format(generator))

        assert 'cmake -G "{}"'.format(generator) in client.out
        assert "GNU-like command-line" in client.out
        assert "main __clang_major__18" in client.out
        assert "main _MSC_VER194" in client.out
        assert "main _MSVC_LANG2014" in client.out
        assert "main _M_X64 defined" in client.out
        assert "main __x86_64__ defined" in client.out
        cmd = re.search(r"MYCMD=(.*)!", str(client.out)).group(1)
        cmd = cmd + ".exe"
        check_vs_runtime(cmd, client, "17", build_type="Release", static_runtime=False)

    @pytest.mark.tool("ninja")
    @pytest.mark.tool("visual_studio", "17")
    def test_clang_cmake_ninja_clang_cl(self, client):
        pr = textwrap.dedent("""\
            [conf]
            tools.build:compiler_executables={"cpp": "clang-cl", "c": "clang-cl", "rc": "clang"}
            """)
        client.save({"comp_exes": pr})

        client.run("create . --name=pkg --version=0.1 -pr=clang -pr=comp_exes "
                   "-s compiler.runtime=dynamic -s compiler.runtime_version=v144 "
                   '-c tools.cmake.cmaketoolchain:generator="Ninja"')

        assert 'cmake -G "Ninja"' in client.out
        assert "MSVC-like command-line" in client.out
        assert "main __clang_major__18" in client.out
        assert "main _MSC_VER194" in client.out
        assert "main _MSVC_LANG2014" in client.out
        assert "main _M_X64 defined" in client.out
        assert "main __x86_64__ defined" in client.out
        cmd = re.search(r"MYCMD=(.*)!", str(client.out)).group(1)
        cmd = cmd + ".exe"
        check_vs_runtime(cmd, client, "17", build_type="Release", static_runtime=False)

    @pytest.mark.tool("ninja")
    @pytest.mark.tool("visual_studio", "17")
    @pytest.mark.tool("clang", "18")  # repeated, for priority over the mingw64 clang
    def test_clang_cmake_runtime_version(self, client):
        generator = "Ninja"
        # Make sure that normal CMakeLists with verify=False works
        client.save({"CMakeLists.txt": gen_cmakelists(verify=False, appname="my_app",
                                                      appsources=["src/main.cpp"], install=True)})
        client.run("create . --name=pkg --version=0.1 -pr=clang -s compiler.runtime=dynamic "
                   "-s compiler.cppstd=17 -s compiler.runtime_version=v144 "
                   '-c tools.cmake.cmaketoolchain:generator="{}"'.format(generator))

        assert 'cmake -G "{}"'.format(generator) in client.out
        assert "GNU-like command-line" in client.out
        assert "main __clang_major__18" in client.out
        # Check this! Clang compiler in Windows is reporting MSC_VER and MSVC_LANG!
        assert "main _MSC_VER194" in client.out
        assert "main _MSVC_LANG2017" in client.out
        assert "main _M_X64 defined" in client.out
        assert "main __x86_64__ defined" in client.out
        cmd = re.search(r"MYCMD=(.*)!", str(client.out)).group(1)
        cmd = cmd + ".exe"
        check_vs_runtime(cmd, client, "17", build_type="Release", static_runtime=False)


@pytest.mark.skipif(platform.system() != "Windows", reason="requires Win")
class TestVSClangCL:
    """
    This is also LLVM/Clang, but distributed with the VS installation
    """
    @pytest.mark.tool("cmake", "3.27")
    @pytest.mark.tool("visual_studio", "17")
    def test_clang_visual_studio_generator(self, client):
        """ This is using the embedded ClangCL compiler, not the external one"""
        generator = "Visual Studio 17"
        client.run("create . --name=pkg --version=0.1 -pr=clang -s compiler.runtime=dynamic "
                   "-s compiler.cppstd=17 -s compiler.runtime_version=v144 "
                   '-c tools.cmake.cmaketoolchain:generator="{}"'.format(generator))
        assert 'cmake -G "{}"'.format(generator) in client.out
        assert "MSVC-like command-line" in client.out
        assert "main __clang_major__19" in client.out
        # Check this! Clang compiler in Windows is reporting MSC_VER and MSVC_LANG!
        # CI forced the installation of 19.38, seems to prevail there
        assert "main _MSC_VER19" in client.out
        assert "main _MSVC_LANG2017" in client.out
        assert "main _M_X64 defined" in client.out
        assert "main __x86_64__ defined" in client.out
        assert "-m64" not in client.out
        cmd = re.search(r"MYCMD=(.*)!", str(client.out)).group(1)
        cmd = cmd + ".exe"
        check_vs_runtime(cmd, client, "17", build_type="Release", static_runtime=False)


@pytest.mark.tool("cmake")
@pytest.mark.skipif(platform.system() != "Windows", reason="requires Win")
class TestMsysClang:
    @pytest.mark.tool("msys2_clang64")
    def test_msys2_clang(self, client):
        """ Using the msys2 clang64 subsystem
        We are not really injecting the msys2 root with make, so using
        MinGW Makefiles
        """
        client.run('create . --name=pkg --version=0.1 -pr=clang -s os.subsystem=msys2 '
                   '-s compiler.libcxx=libc++ '
                   '-c tools.cmake.cmaketoolchain:generator="MinGW Makefiles"')
        # clang compilations in Windows will use MinGW Makefiles by default
        assert 'cmake -G "MinGW Makefiles"' in client.out
        # TODO: Version is still not controlled
        assert "main __clang_major__19" in client.out
        # Not using libstdc++
        assert "_GLIBCXX_USE_CXX11_ABI" not in client.out
        assert "main __cplusplus2014" in client.out
        assert "main __GNUC__" in client.out
        assert "main __MINGW32__1" in client.out
        assert "main __MINGW64__1" in client.out
        assert "main _MSC_" not in client.out
        assert "main _MSVC_" not in client.out
        assert "main _M_X64 defined" in client.out
        assert "main __x86_64__ defined" in client.out

        cmd = re.search(r"MYCMD=(.*)!", str(client.out)).group(1)
        cmd = cmd + ".exe"
        check_vs_runtime(cmd, client, "17", build_type="Release",
                         static_runtime=False, subsystem="clang64")

    @pytest.mark.tool("msys2_mingw64_clang64")
    def test_msys2_clang_mingw(self, client):
        """ compiling with the clang INSIDE mingw, which uses the
        MinGW runtime, not the MSVC one
        For 32 bits, it doesn't seem possible to install the toolchain
        For 64 bits require "pacman -S mingw-w64-x86-clang++"
        """
        # TODO: This should probably go to the ``os.subsystem=ming64" but lets do it in other PR
        client.run('create . --name=pkg --version=0.1 -pr=clang '
                   '-s compiler.libcxx=libstdc++')
        # clang compilations in Windows will use MinGW Makefiles by default
        assert 'cmake -G "MinGW Makefiles"' in client.out
        # TODO: Version is still not controlled
        assert "main __clang_major__19" in client.out
        assert "main _GLIBCXX_USE_CXX11_ABI 0" in client.out
        assert "main __cplusplus2014" in client.out
        assert "main __GNUC__" in client.out
        assert "main __MINGW32__1" in client.out
        assert "main __MINGW64__1" in client.out
        assert "main _MSC_" not in client.out
        assert "main _MSVC_" not in client.out
        assert "main _M_X64 defined" in client.out
        assert "main __x86_64__ defined" in client.out

        cmd = re.search(r"MYCMD=(.*)!", str(client.out)).group(1)
        cmd = cmd + ".exe"
        check_vs_runtime(cmd, client, "17", build_type="Release",
                         static_runtime=False, subsystem="mingw64")

    @pytest.mark.tool("msys2_clang64")
    def test_clang_pure_c(self, client):
        """ compiling with the clang INSIDE mingw, which uses the
        MinGW runtime, not the MSVC one
        For 32 bits, it doesn't seem possible to install the toolchain
        For 64 bits require "pacman -S mingw-w64-x86-clang++"
        """
        client.save({"CMakeLists.txt": gen_cmakelists(verify=False, language="C", appname="my_app",
                                                      appsources=["src/main.c"], install=True),
                     "src/main.c": gen_function_c(name="main")})
        client.run(f"create . --name=pkg --version=0.1 -pr=clang")
        # clang compilations in Windows will use MinGW Makefiles by default
        assert 'cmake -G "MinGW Makefiles"' in client.out
        assert "main __clang_major__19" in client.out
        assert "GLIBCXX" not in client.out
        assert "cplusplus" not in client.out
        assert "main __GNUC__" in client.out
        assert "main __MINGW32__1" in client.out
        assert "main __MINGW64__1" in client.out
        assert "main _MSC_" not in client.out
        assert "main _MSVC_" not in client.out
        assert "main _M_X64 defined" in client.out
        assert "main __x86_64__ defined" in client.out

        cmd = re.search(r"MYCMD=(.*)!", str(client.out)).group(1)
        cmd = cmd + ".exe"
        # static_runtime equivalent to C, for checking, no dep on libc++
        check_vs_runtime(cmd, client, "17", build_type="Release", static_runtime=True,
                         subsystem="clang64")


@pytest.mark.tool("cmake")
@pytest.mark.tool("ninja")
@pytest.mark.tool("clang", "18")
@pytest.mark.skipif(platform.system() != "Windows", reason="requires Win")
def test_error_clang_cmake_ninja_custom_cxx(client):
    clang_profile = textwrap.dedent("""
        [settings]
        os=Windows
        arch=x86_64
        build_type=Release
        compiler=clang
        compiler.version=18
        [buildenv]
        CXX=/no/exist/clang++
        """)
    client.save({"clang":     clang_profile})
    client.run("create . --name=pkg --version=0.1 -pr=clang "
               "-c tools.cmake.cmaketoolchain:generator=Ninja", assert_error=True)
    assert 'Could not find compiler' in client.out
    assert '/no/exist/clang++' in client.out



import json

import yaml

from conan.api.model import RecipeReference
from conan.errors import ConanException
from conan.internal.util.files import load, save


def loadconanconfig(filename):
    try:
        contents = json.loads(load(filename))
        config_versions = contents["config_version"]
        config_versions = [RecipeReference.loads(r) for r in config_versions]
    except Exception as e:
        raise ConanException(f"Error while loading config file {filename}: {str(e)}")
    return config_versions


def loadconanconfig_yml(filename):
    try:
        contents = yaml.safe_load(load(filename))
        config_versions = contents["packages"]
        config_versions = [RecipeReference.loads(r) for r in config_versions]
        urls = contents.get("urls")
    except Exception as e:
        raise ConanException(f"Error while loading config file {filename}: {str(e)}")
    return config_versions, urls


def saveconanconfig(filename, config_versions):
    try:
        config_versions = [r.repr_notime() for r in config_versions]
        save(filename, json.dumps({"config_version": config_versions}, indent=4))
    except Exception as e:
        raise ConanException(f"Error while saving config file {filename}: {str(e)}")



import os
import textwrap

import pytest
import yaml

from conan.api.conan_api import ConanAPI
from conan.internal.cache.cache import PkgCache
from conan.internal.cache.home_paths import HomePaths
from conan.internal.default_settings import default_settings_yml
from conan.internal.model.conf import ConfDefinition
from conan.internal.model.manifest import FileTreeManifest
from conan.internal.model.options import Options
from conan.internal.model.profile import Profile
from conan.api.model import RecipeReference
from conan.internal.model.settings import Settings
from conan.test.utils.test_files import temp_folder
from conan.test.utils.tools import GenConanfile
from conan.internal.util.dates import revision_timestamp_now
from conan.internal.util.files import save


class GraphManagerTest:

    @pytest.fixture(autouse=True)
    def setUp(self):
        cache_folder = temp_folder()
        cache = PkgCache(cache_folder, ConfDefinition())
        home = HomePaths(cache_folder)
        save(os.path.join(home.profiles_path, "default"), "")
        save(home.settings_path, "os: [Windows, Linux]")
        self.cache = cache
        self.cache_folder = cache_folder

    def recipe_cache(self, reference, requires=None, option_shared=None):
        ref = RecipeReference.loads(reference)
        conanfile = GenConanfile()
        if requires:
            for r in requires:
                conanfile.with_require(r)
        if option_shared is not None:
            conanfile.with_option("shared", [True, False])
            conanfile.with_default_option("shared", option_shared)

        self._cache_recipe(ref, conanfile)

    def recipe_conanfile(self, reference, conanfile):
        ref = RecipeReference.loads(reference)
        self._cache_recipe(ref, conanfile)

    def _cache_recipe(self, ref, test_conanfile):
        if not isinstance(ref, RecipeReference):
            ref = RecipeReference.loads(ref)
        ref.revision = "123"
        ref.timestamp = revision_timestamp_now()
        recipe_layout = self.cache.create_ref_layout(ref)
        save(recipe_layout.conanfile(), str(test_conanfile))
        manifest = FileTreeManifest.create(recipe_layout.export())
        manifest.save(recipe_layout.export())

    def alias_cache(self, alias, target):
        ref = RecipeReference.loads(alias)
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class Alias(ConanFile):
                alias = "%s"
            """ % target)
        self._cache_recipe(ref, conanfile)

    @staticmethod
    def recipe_consumer(reference=None, requires=None, build_requires=None):
        path = temp_folder()
        path = os.path.join(path, "conanfile.py")
        conanfile = GenConanfile()
        if reference:
            ref = RecipeReference.loads(reference)
            conanfile.with_name(ref.name).with_version(ref.version)
        if requires:
            for r in requires:
                conanfile.with_require(r)
        if build_requires:
            for r in build_requires:
                conanfile.with_build_requires(r)
        save(path, str(conanfile))
        return path

    @staticmethod
    def consumer_conanfile(conanfile):
        path = temp_folder()
        path = os.path.join(path, "conanfile.py")
        save(path, str(conanfile))
        return path

    def build_graph(self, content, profile_build_requires=None, install=True, options_build=None):
        path = temp_folder()
        path = os.path.join(path, "conanfile.py")
        save(path, str(content))
        return self.build_consumer(path, profile_build_requires, install,
                                   options_build=options_build)

    def build_consumer(self, path, profile_build_requires=None, install=True, options_build=None):
        profile_host = Profile()
        profile_host.settings["os"] = "Linux"
        profile_build = Profile()
        profile_build.settings["os"] = "Windows"
        if profile_build_requires:
            profile_host.tool_requires = profile_build_requires
        if options_build:
            profile_build.options = Options(options_values=options_build)
        cache_settings = Settings(yaml.safe_load(default_settings_yml))
        profile_host.process_settings(cache_settings)
        profile_build.process_settings(cache_settings)
        build_mode = ["*"]  # Means build all

        conan_api = ConanAPI(cache_folder=self.cache_folder)

        deps_graph = conan_api.graph.load_graph_consumer(path, None, None, None, None,
                                                         profile_host, profile_build, None, None,
                                                         None)

        if install:
            deps_graph.report_graph_error()
            conan_api.graph.analyze_binaries(deps_graph, build_mode)
            conan_api.install.install_binaries(deps_graph)

        return deps_graph

    @staticmethod
    def _check_node(node, ref, deps=None, dependents=None, settings=None, options=None):
        dependents = dependents or []
        deps = deps or []

        conanfile = node.conanfile
        ref = RecipeReference.loads(str(ref))
        assert node.ref == ref
        if conanfile:
            assert conanfile.name == ref.name

        assert len(node.edges) == len(deps)
        for d in node.neighbors():
            assert d in deps

        dependants = node.inverse_neighbors()
        assert len(dependants) == len(dependents)
        for d in dependents:
            assert d in dependants

        if settings is not None:
            for k, v in settings.items():
                assert conanfile.settings.get_safe(k) == v

        if options is not None:
            for k, v in options.items():
                assert conanfile.options.get_safe(k) == v



import copy
import os
import re
from fnmatch import translate

from conan.internal.errors import ForbiddenException, RecipeNotFoundException
from conan.api.model import PkgReference
from conan.api.model import RecipeReference
from conan.internal.paths import CONANINFO
from conans.server.utils.files import list_folder_subdirs
from conan.internal.util.files import load


def _get_local_infos_min(server_store, ref, list_only):
    result = {}
    new_ref = ref
    subdirs = list_folder_subdirs(server_store.packages(new_ref), level=1)
    # Seems that Python3.12 os.walk is retrieving a different order of folders
    for package_id in sorted(subdirs):
        if package_id in result:
            continue
        if list_only:
            result[package_id] = {}
            continue
        # Read conaninfo
        pref = PkgReference(new_ref, package_id)
        revision_entry = server_store.get_last_package_revision(pref)
        if not revision_entry:
            continue  # server can store empty package-revision list files
        pref.revision = revision_entry.revision
        info_path = os.path.join(server_store.package(pref), CONANINFO)
        if not os.path.exists(info_path):
            raise Exception(f"No conaninfo.txt file for listed {pref}")
        content = load(info_path)
        # From Conan 1.48 the conaninfo.txt is sent raw.
        result[package_id] = {"content": content}

    return result


def search_packages(server_store, ref, list_only):
    """
    Return a dict like this:

            {package_ID: {name: "OpenCV",
                           version: "2.14",
                           settings: {os: Windows}}}
    param ref: RecipeReference object
    """
    if ref.revision is None:  # TODO: Verify this sometimes happen
        latest_rev = server_store.get_last_revision(ref).revision
        ref.revision = latest_rev

    ref_norev = copy.copy(ref)
    ref_norev.revision = None
    if not os.path.exists(server_store.conan_revisions_root(ref_norev)):
        raise RecipeNotFoundException(ref)
    infos = _get_local_infos_min(server_store, ref, list_only)
    return infos


class SearchService(object):

    def __init__(self, authorizer, server_store, auth_user):
        self._authorizer = authorizer
        self._server_store = server_store
        self._auth_user = auth_user

    def search_packages(self, reference, list_only=False):
        """Shared between v1 and v2, v1 will iterate rrevs"""
        self._authorizer.check_read_conan(self._auth_user, reference)
        info = search_packages(self._server_store, reference, list_only)
        return info

    def _search_recipes(self, pattern=None, ignorecase=True):
        subdirs = list_folder_subdirs(basedir=self._server_store.store, level=5)

        def underscore_to_none(field):
            return field if field != "_" else None

        ret = set()
        if not pattern:
            for folder in subdirs:
                fields_dir = [underscore_to_none(d) for d in folder.split("/")]
                r = RecipeReference(*fields_dir)
                r.revision = None
                ret.add(r)
        else:
            # Conan references in main storage
            pattern = str(pattern)
            b_pattern = translate(pattern)
            b_pattern = re.compile(b_pattern, re.IGNORECASE) if ignorecase else re.compile(b_pattern)
            for subdir in subdirs:
                fields_dir = [underscore_to_none(d) for d in subdir.split("/")]
                new_ref = RecipeReference(*fields_dir)
                new_ref.revision = None
                if new_ref.partial_match(b_pattern):
                    ret.add(new_ref)

        return sorted(ret)

    def search(self, pattern=None, ignorecase=True):
        """ Get all the info about any package
            Attributes:
                pattern = wildcards like opencv/*
        """
        refs = self._search_recipes(pattern, ignorecase)
        filtered = []
        # Filter out restricted items
        for ref in refs:
            try:
                self._authorizer.check_read_conan(self._auth_user, ref)
                filtered.append(ref)
            except ForbiddenException:
                pass
        return filtered



import json
import textwrap

import pytest

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


class TestFilterProfile:
    @pytest.fixture()
    def client(self):
        c = TestClient()
        c.save({"lib/conanfile.py": GenConanfile("lib", "1.0").with_settings("os", "build_type")
               .with_shared_option()})
        c.run("create lib -s os=Linux")
        c.run("create lib -s os=Windows")
        c.run("create lib -s os=Windows -o *:shared=True")
        return c

    def test_exact_match(self, client):
        c = client
        c.run("graph explain --requires=lib/1.0 --missing=lib/1.0 -s os=Windows")
        expected = textwrap.dedent("""\
            settings: Windows, Release
            options: shared=False
            diff
              explanation: This binary is an exact match for the defined inputs
            """)
        assert textwrap.indent(expected, "      ") in c.out
        c.run("graph explain --requires=lib/1.0 --missing=lib/1.0 -s os=Windows --format=json")
        closest = json.loads(c.stdout)["closest_binaries"]
        revisions = closest["lib/1.0"]["revisions"]
        pkgs = revisions["5313a980ea0c56baeb582c510d6d9fbc"]["packages"]
        assert len(pkgs) == 1
        pkg1 = pkgs["3d714b452400b3c3d6a964f42d5ec5004a6f22dc"]
        assert pkg1["diff"]["platform"] == {}
        assert pkg1["diff"]["settings"] == {}
        assert pkg1["diff"]["options"] == {}
        assert pkg1["diff"]["dependencies"] == {}
        assert pkg1["diff"]["explanation"] == "This binary is an exact match for the defined inputs"

    def test_settings_incomplete(self, client):
        c = client
        c.save({"windows": "[settings]\nos=Windows"})
        c.run("graph explain --requires=lib/1.0 -pr windows")
        expected = textwrap.dedent("""\
            settings: Windows, Release
            options: shared=False
            diff
              settings
                expected: build_type=None
                existing: build_type=Release
              explanation: This binary was built with different settings.
            """)
        assert textwrap.indent(expected, "      ") in c.out
        c.run("graph explain --requires=lib/1.0 -pr windows --format=json")
        closest = json.loads(c.stdout)["closest_binaries"]
        revisions = closest["lib/1.0"]["revisions"]
        pkgs = revisions["5313a980ea0c56baeb582c510d6d9fbc"]["packages"]
        assert len(pkgs) == 1
        pkg1 = pkgs["3d714b452400b3c3d6a964f42d5ec5004a6f22dc"]
        assert pkg1["diff"]["platform"] == {}
        assert pkg1["diff"]["settings"] == {'existing': ['build_type=Release'],
                                            'expected': ['build_type=None']}
        assert pkg1["diff"]["options"] == {}
        assert pkg1["diff"]["dependencies"] == {}
        assert pkg1["diff"]["explanation"] == "This binary was built with different settings."

    def test_settings_with_option(self, client):
        c = client
        c.save({"windows": "[settings]\nos=Windows\n[options]\n*:shared=True"})
        c.run("graph explain --requires=lib/1.0 -pr windows")
        expected = textwrap.dedent("""\
            settings: Windows, Release
            options: shared=True
            diff
              settings
                expected: build_type=None
                existing: build_type=Release
              explanation: This binary was built with different settings.
            """)
        assert textwrap.indent(expected, "      ") in c.out

    def test_different_option(self, client):
        # We find 1 closest match in Linux static
        c = client
        c.save({"linux": "[settings]\nos=Linux\nbuild_type=Release\n[options]\n*:shared=True"})
        c.run("graph explain --requires=lib/1.0 -pr linux")
        expected = textwrap.dedent("""\
            remote: Local Cache
            settings: Linux, Release
            options: shared=False
            diff
              options
                expected: shared=True
                existing: shared=False
              explanation: This binary was built with the same settings, but different options
            """)
        assert textwrap.indent(expected, "      ") in c.out
        c.run("graph explain --requires=lib/1.0 -pr linux --format=json")
        closest = json.loads(c.stdout)["closest_binaries"]
        revisions = closest["lib/1.0"]["revisions"]
        pkgs = revisions["5313a980ea0c56baeb582c510d6d9fbc"]["packages"]
        assert len(pkgs) == 1
        pkg1 = pkgs["499989797d9192081b8f16f7d797b107a2edd8da"]
        assert pkg1["diff"]["platform"] == {}
        assert pkg1["diff"]["settings"] == {}
        assert pkg1["diff"]["options"] == {'existing': ['shared=False'], 'expected': ['shared=True']}
        assert pkg1["diff"]["dependencies"] == {}
        assert pkg1["diff"]["explanation"] == "This binary was built with the same settings, " \
                                              "but different options"

    def test_different_option_none(self):
        # We find 1 closest match in Linux static
        c = TestClient()
        c.save({"conanfile.py": GenConanfile("lib", "1.0").with_option("opt", [None, 1])})
        c.run("create .")
        c.run("graph explain --requires=lib/1.0 -o *:opt=1")
        expected = textwrap.dedent("""\
            remote: Local Cache
            diff
              options
                expected: opt=1
                existing: opt=None
              explanation: This binary was built with the same settings, but different options
            """)
        assert textwrap.indent(expected, "      ") in c.out

        c.run("remove * -c")
        c.run("create . -o *:opt=1")
        c.run("graph explain --requires=lib/1.0")
        expected = textwrap.dedent("""\
            remote: Local Cache
            options: opt=1
            diff
              options
                expected: opt=None
                existing: opt=1
              explanation: This binary was built with the same settings, but different options
            """)
        assert textwrap.indent(expected, "      ") in c.out

    def test_different_setting(self, client):
        # We find 1 closest match in Linux static
        c = client
        c.save({"windows": "[settings]\nos=Windows\nbuild_type=Debug\n[options]\n*:shared=True"})
        c.run("graph explain --requires=lib/1.0 -pr windows")
        expected = textwrap.dedent("""\
            remote: Local Cache
            settings: Windows, Release
            options: shared=True
            diff
              settings
                expected: build_type=Debug
                existing: build_type=Release
              explanation: This binary was built with different settings.
            """)
        assert textwrap.indent(expected, "      ") in c.out
        c.run("graph explain --requires=lib/1.0 -pr windows --format=json")
        closest = json.loads(c.stdout)["closest_binaries"]
        revisions = closest["lib/1.0"]["revisions"]
        pkgs = revisions["5313a980ea0c56baeb582c510d6d9fbc"]["packages"]
        assert len(pkgs) == 1
        pkg1 = pkgs["c2dd2d51b5074bdb5b7d717929372de09830017b"]
        assert pkg1["diff"]["platform"] == {}
        assert pkg1["diff"]["settings"] == {'existing': ['build_type=Release'],
                                            'expected': ['build_type=Debug']}
        assert pkg1["diff"]["options"] == {}
        assert pkg1["diff"]["dependencies"] == {}
        assert pkg1["diff"]["explanation"] == "This binary was built with different settings."

    def test_different_settings_target(self):
        c = TestClient()
        conanfile = textwrap.dedent("""\
            from conan import ConanFile
            class Pkg(ConanFile):
                name = "tool"
                version = "1.0"
                def package_id(self):
                    self.info.settings_target = self.settings_target.copy()
                    self.info.settings_target.constrained(["os"])
                """)
        c.save({"conanfile.py": conanfile})
        c.run("create . --build-require -s:b os=Windows -s:h os=Linux")

        c.run("graph explain --tool-requires=tool/1.0 -s:b os=Windows -s:h os=Macos")
        expected = textwrap.dedent("""\
            remote: Local Cache
            settings_target: os=Linux
            diff
              settings_target
                expected: os=Macos
                existing: os=Linux
              explanation: This binary was built with different settings_target.
            """)
        assert textwrap.indent(expected, "      ") in c.out
        c.run("graph explain --tool-requires=tool/1.0 -s:b os=Windows -s:h os=Macos --format=json")
        closest = json.loads(c.stdout)["closest_binaries"]
        revisions = closest["tool/1.0"]["revisions"]
        pkgs = revisions["4cc4b286a46dc2ed188d8c417eadb4e6"]["packages"]
        assert len(pkgs) == 1
        pkg1 = pkgs["d66135125c07cc240b8d6adda090b76d60341205"]
        assert pkg1["diff"]["platform"] == {}
        assert pkg1["diff"]["settings_target"] == {'existing': ['os=Linux'],
                                                   'expected': ['os=Macos']}
        assert pkg1["diff"]["options"] == {}
        assert pkg1["diff"]["dependencies"] == {}
        assert pkg1["diff"]["explanation"] == "This binary was built with different settings_target."

    def test_different_platform(self, client):
        # We find closest match in other platforms
        c = client
        c.save({"macos": "[settings]\nos=Macos\nbuild_type=Release\n[options]\n*:shared=True"})
        c.run("graph explain --requires=lib/1.0 -pr macos")
        expected = textwrap.dedent("""\
            remote: Local Cache
            settings: Windows, Release
            options: shared=True
            diff
              platform
                expected: os=Macos
                existing: os=Windows
              explanation: This binary belongs to another OS or Architecture, highly incompatible.
            """)
        assert textwrap.indent(expected, "      ") in c.out
        c.run("graph explain --requires=lib/1.0 -pr macos --format=json")
        closest = json.loads(c.stdout)["closest_binaries"]
        revisions = closest["lib/1.0"]["revisions"]
        pkgs = revisions["5313a980ea0c56baeb582c510d6d9fbc"]["packages"]
        assert len(pkgs) == 1
        pkg1 = pkgs["c2dd2d51b5074bdb5b7d717929372de09830017b"]
        assert pkg1["diff"]["platform"] == {'existing': ['os=Windows'], 'expected': ['os=Macos']}
        assert pkg1["diff"]["settings"] == {}
        assert pkg1["diff"]["options"] == {}
        assert pkg1["diff"]["dependencies"] == {}
        assert pkg1["diff"]["explanation"] == "This binary belongs to another OS or Architecture, " \
                                              "highly incompatible."

    def test_different_conf(self, client):
        # We find closest match in other platforms
        c = client
        c.save_home({"global.conf": "tools.info.package_id:confs=['user.foo:bar']"})
        c.run("graph explain --requires=lib/1.0 -c user.foo:bar=42 -s os=Linux")
        expected = textwrap.dedent("""\
            remote: Local Cache
            settings: Linux, Release
            options: shared=False
            diff
              confs
                expected: user.foo:bar=42
                existing: user.foo:bar=None
              explanation: This binary has same settings, options and dependencies, but different confs
            """)
        assert textwrap.indent(expected, "      ") in c.out
        c.run("graph explain --requires=lib/1.0 -c user.foo:bar=42 -s os=Linux -f=json")
        closest = json.loads(c.stdout)["closest_binaries"]
        revisions = closest["lib/1.0"]["revisions"]
        pkgs = revisions["5313a980ea0c56baeb582c510d6d9fbc"]["packages"]
        assert len(pkgs) == 1
        pkg1 = pkgs["499989797d9192081b8f16f7d797b107a2edd8da"]
        assert pkg1["diff"]["platform"] == {}
        assert pkg1["diff"]["settings"] == {}
        assert pkg1["diff"]["options"] == {}
        assert pkg1["diff"]["dependencies"] == {}
        assert pkg1["diff"]["confs"] == {"expected": ["user.foo:bar=42"],
                                         "existing": ["user.foo:bar=None"]}
        assert pkg1["diff"]["explanation"] == "This binary has same settings, options and " \
                                              "dependencies, but different confs"


class TestMissingBinaryDeps:
    @pytest.fixture()
    def client(self):
        c = TestClient()
        c.save({"dep/conanfile.py": GenConanfile("dep").with_settings("os"),
                "lib/conanfile.py": GenConanfile("lib", "1.0").with_settings("os")
               .with_requires("dep/[>=1.0]")})
        c.run("create dep --version=1.0 -s os=Linux")
        c.run("create lib -s os=Linux")
        c.run("create dep --version=2.0 -s os=Linux")
        return c

    def test_other_platform(self, client):
        c = client
        c.run("install --requires=lib/1.0 -s os=Windows", assert_error=True)
        assert "ERROR: Missing prebuilt package for 'dep/2.0', 'lib/1.0'" in c.out
        # We use the --missing=lib/1.0 to specify we want this binary and not dep/2.0
        c.run("graph explain --requires=lib/1.0 --missing=lib/1.0 -s os=Windows")
        expected = textwrap.dedent("""\
            settings: Linux
            requires: dep/1.Y.Z
            diff
              platform
                expected: os=Windows
                existing: os=Linux
              dependencies
                expected: dep/2.Y.Z
                existing: dep/1.Y.Z
              explanation: This binary belongs to another OS or Architecture, highly incompatible.
            """)
        assert textwrap.indent(expected, "      ") in c.out

    def test_other_dependencies(self, client):
        c = client
        c.run("install --requires=lib/1.0 -s os=Linux", assert_error=True)
        assert "ERROR: Missing prebuilt package for 'lib/1.0'" in c.out
        c.run("graph explain --requires=lib/1.0 -s os=Linux")
        expected = textwrap.dedent("""\
            settings: Linux
            requires: dep/1.Y.Z
            diff
              dependencies
                expected: dep/2.Y.Z
                existing: dep/1.Y.Z
              explanation: This binary has same settings and options, but different dependencies
               """)
        assert textwrap.indent(expected, "      ") in c.out

    def test_different_python_requires(self):
        c = TestClient(light=True)
        c.save({"tool/conanfile.py": GenConanfile("tool"),
                "lib/conanfile.py": GenConanfile("lib", "1.0").with_python_requires("tool/[>=1.0]")})
        c.run("create tool --version=1.0")
        c.run("create lib")
        c.run("create tool --version=2.0")
        c.run("graph explain --requires=lib/1.0")
        expected = textwrap.dedent("""\
            remote: Local Cache
            python_requires: tool/1.0.Z
            diff
              python_requires
                expected: tool/2.0.Z
                existing: tool/1.0.Z
              explanation: This binary has same settings, options and dependencies, but different python_requires
            """)
        assert textwrap.indent(expected, "      ") in c.out
        c.run("graph explain --requires=lib/1.0 -s os=Linux --format=json")
        closest = json.loads(c.stdout)["closest_binaries"]
        revisions = closest["lib/1.0"]["revisions"]
        pkgs = revisions["7bf17caa5bf9d2ed1dd8b337e9623fc0"]["packages"]
        assert len(pkgs) == 1
        pkg1 = pkgs["5ccdb706197ca94edc0ecee9ef0d0b11b887d937"]
        assert pkg1["diff"]["platform"] == {}
        assert pkg1["diff"]["settings"] == {}
        assert pkg1["diff"]["options"] == {}
        assert pkg1["diff"]["dependencies"] == {}
        assert pkg1["diff"]["python_requires"] == {"expected": ["tool/2.0.Z"],
                                                   "existing": ["tool/1.0.Z"]}
        assert pkg1["diff"]["explanation"] == "This binary has same settings, options and " \
                                              "dependencies, but different python_requires"

    def test_build_requires(self):
        c = TestClient(light=True)
        c.save_home({"global.conf": "core.package_id:default_build_mode=minor_mode"})
        c.save({"tool/conanfile.py": GenConanfile("tool"),
                "lib/conanfile.py": GenConanfile("lib", "1.0").with_tool_requires("tool/[>=1.0]")})
        c.run("create tool --version=1.0")
        c.run("create lib")
        c.run("create tool --version=2.0")
        c.run("graph explain --requires=lib/1.0")
        expected = textwrap.dedent("""\
            remote: Local Cache
            build_requires: tool/1.0.Z
            diff
              build_requires
                expected: tool/2.0.Z
                existing: tool/1.0.Z
              explanation: This binary has same settings, options and dependencies, but different build_requires
            """)
        assert textwrap.indent(expected, "      ") in c.out
        c.run("graph explain --requires=lib/1.0 -s os=Linux --format=json")
        closest = json.loads(c.stdout)["closest_binaries"]
        revisions = closest["lib/1.0"]["revisions"]
        pkgs = revisions["4790f7f1561b52be5d39f0bc8e9acbed"]["packages"]
        assert len(pkgs) == 1
        pkg1 = pkgs["c9d96a611b8c819f35728d58d743f6a78a1b5942"]
        assert pkg1["diff"]["platform"] == {}
        assert pkg1["diff"]["settings"] == {}
        assert pkg1["diff"]["options"] == {}
        assert pkg1["diff"]["dependencies"] == {}
        assert pkg1["diff"]["build_requires"] == {"expected": ["tool/2.0.Z"],
                                                  "existing": ["tool/1.0.Z"]}
        assert pkg1["diff"]["explanation"] == "This binary has same settings, options and " \
                                              "dependencies, but different build_requires"


def test_change_in_package_type():
    tc = TestClient(light=True)
    tc.save({
        "libc/conanfile.py": GenConanfile("libc", "1.0"),
        "libb/conanfile.py": GenConanfile("libb", "1.0")
        .with_requires("libc/1.0"),
        "liba/conanfile.py": GenConanfile("liba", "1.0")
        .with_requires("libb/1.0")
    })

    tc.run("create libc")
    tc.run("create libb")
    tc.run("create liba")

    tc.save({
        "libc/conanfile.py": GenConanfile("libc", "1.0")
        .with_package_type("application")
    })
    tc.run("create libc")

    tc.run("create liba", assert_error=True)
    assert "Missing binary: libb/1.0" in tc.out

    tc.run("graph explain --requires=liba/1.0")
    # This fails, graph explain thinks everything is ok
    assert "explanation: This binary is an exact match for the defined inputs" not in tc.out


def test_conf_difference_shown():
    tc = TestClient(light=True)
    tc.save({
        "libc/conanfile.py": GenConanfile("libc", "1.0"),
        "libb/conanfile.py": GenConanfile("libb", "1.0").with_requires("libc/1.0"),
        "liba/conanfile.py": GenConanfile("liba", "1.0").with_requires("libb/1.0")
    })
    tc.save_home({"global.conf": "tools.info.package_id:confs=['user.foo:bar']"})

    tc.run("create libc")
    tc.run("create libb")
    tc.run("create liba")

    tc.run("remove libc/*:* -c")

    tc.run("create libc -c user.foo:bar=42")

    tc.run("create liba", assert_error=True)
    assert "Missing prebuilt package for 'libc/1.0'" in tc.out

    tc.run("graph explain --requires=liba/1.0")
    assert "conf: user.foo:bar=42" in tc.out


class TestDistance:
    def test_multiple_distance_ordering(self):
        tc = TestClient()
        tc.save({
            "conanfile.py": GenConanfile("pkg", "1.0").with_requires("dep/1.0"),
            "dep/conanfile.py": GenConanfile("dep", "1.0")
            .with_option("shared", [True, False])
            .with_option("fPIC", [True, False])})

        tc.run("create dep -o shared=True -o fPIC=True")
        tc.run("create dep -o shared=True -o fPIC=False")

        tc.run('graph explain . -o *:shared=False -o *:fPIC=False')
        # We don't expect the further binary to show
        assert "a657a8fc96dd855e2a1c90a9fe80125f0c4635a0" not in tc.out
        # We expect the closer binary to show
        assert "a6923b987deb1469815dda84aab36ac34f370c48" in tc.out


def test_no_binaries():
    # https://github.com/conan-io/conan/issues/15819
    c = TestClient()
    c.save({"conanfile.py": GenConanfile("pkg", "0.1")})
    c.run("export .")
    c.run("graph explain --requires=pkg/0.1")
    assert "ERROR: No package binaries exist" in c.out
    # The json is not an issue, it won't have anything as contents



import textwrap

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


class TestTestRequiresDiamond:
    def test_test_requires_linear(self):
        c = TestClient(light=True)
        c.save({"zlib/conanfile.py": GenConanfile("zlib", "1.0"),
                "gtest/conanfile.py": GenConanfile("gtest", "1.0").with_requires("zlib/1.0"),
                "engine/conanfile.py": GenConanfile("engine", "1.0").with_test_requires("gtest/1.0")
                })
        c.run("create zlib")
        c.run("create gtest")
        c.run("install engine")
        c.assert_listed_require({"gtest/1.0": "Cache",
                                 "zlib/1.0": "Cache"}, test=True)

    def test_test_requires_half_diamond(self):
        c = TestClient(light=True)
        c.save({"zlib/conanfile.py": GenConanfile("zlib", "1.0"),
                "gtest/conanfile.py": GenConanfile("gtest", "1.0").with_requires("zlib/1.0"),
                "engine/conanfile.py": GenConanfile("engine", "1.0").with_requires("zlib/1.0")
                                                                    .with_test_requires("gtest/1.0")
                })
        c.run("create zlib")
        c.run("create gtest")
        c.run("install engine")
        c.assert_listed_require({"zlib/1.0": "Cache"})
        c.assert_listed_require({"gtest/1.0": "Cache"}, test=True)

    def test_test_requires_half_diamond_change_order(self):
        engine = textwrap.dedent("""
            from conan import ConanFile
            class Pkg(ConanFile):
                def requirements(self):
                    # Best practice is to declare test_requires in the build_requirements() method,
                    # but this ensures it is also possible to declare them in requirements(),
                    # as has historically been allowed.
                    self.test_requires("gtest/1.0")
                    self.requires("zlib/1.0")
            """)
        c = TestClient(light=True)
        c.save({"zlib/conanfile.py": GenConanfile("zlib", "1.0"),
                "gtest/conanfile.py": GenConanfile("gtest", "1.0").with_requires("zlib/1.0"),
                "engine/conanfile.py": engine
                })
        c.run("create zlib")
        c.run("create gtest")
        c.run("install engine")
        c.assert_listed_require({"zlib/1.0": "Cache"})
        c.assert_listed_require({"gtest/1.0": "Cache"}, test=True)

    def test_test_requires_diamond(self):
        c = TestClient(light=True)
        c.save({"zlib/conanfile.py": GenConanfile("zlib", "1.0"),
                "gtest/conanfile.py": GenConanfile("gtest", "1.0").with_requires("zlib/1.0"),
                "engine/conanfile.py": GenConanfile("engine", "1.0").with_requires("zlib/1.0"),
                "game/conanfile.py": GenConanfile().with_requires("engine/1.0")
                                                   .with_test_requires("gtest/1.0")
                })
        c.run("create zlib")
        c.run("create gtest")
        c.run("create engine")
        c.run("install game")
        c.assert_listed_require({"zlib/1.0": "Cache",
                                 "engine/1.0": "Cache"})
        c.assert_listed_require({"gtest/1.0": "Cache"}, test=True)

    def test_test_requires_diamond_change_order(self):
        c = TestClient(light=True)
        game = textwrap.dedent("""
           from conan import ConanFile
           class Pkg(ConanFile):
               def build_requirements(self):
                   self.test_requires("gtest/1.0")
               def requirements(self):
                   self.requires("engine/1.0")
           """)
        c.save({"zlib/conanfile.py": GenConanfile("zlib", "1.0"),
                "gtest/conanfile.py": GenConanfile("gtest", "1.0").with_requires("zlib/1.0"),
                "engine/conanfile.py": GenConanfile("engine", "1.0").with_requires("zlib/1.0"),
                "game/conanfile.py": game
                })
        c.run("create zlib")
        c.run("create gtest")
        c.run("create engine")
        c.run("install game")
        c.assert_listed_require({"zlib/1.0": "Cache",
                                 "engine/1.0": "Cache"})
        c.assert_listed_require({"gtest/1.0": "Cache"}, test=True)

    def test_test_requires_conflict_force(self):
        c = TestClient(light=True)
        game = textwrap.dedent("""
           from conan import ConanFile
           class Pkg(ConanFile):
               def build_requirements(self):
                   self.test_requires("gtest/1.0", force=True)
                   self.test_requires("rapidcheck/1.0")
           """)
        c.save({"gtest/conanfile.py": GenConanfile("gtest"),
                "rapidcheck/conanfile.py":
                    GenConanfile("rapidcheck", "1.0").with_requires("gtest/1.1"),
                "game/conanfile.py": game
                })
        c.run("create gtest --version=1.0")
        c.run("create gtest --version=1.1")
        c.run("create rapidcheck")
        c.run("install game")
        c.assert_listed_require({"gtest/1.0": "Cache"}, test=True)
        c.assert_overrides({"gtest/1.1": ["gtest/1.0"]})


def test_require_options():
    c = TestClient(light=True)
    gtest = textwrap.dedent("""
        from conan import ConanFile
        class Gtest(ConanFile):
            name = "gtest"
            version = "1.0"
            options = {"myoption": [1, 2, 3]}
            default_options = {"myoption": 1}
            def package_info(self):
                self.output.info(f"MYOPTION: {self.options.myoption}")
        """)
    engine = textwrap.dedent("""
        from conan import ConanFile
        class Engine(ConanFile):
            name = "engine"
            version = "1.0"
            def build_requirements(self):
                self.test_requires("gtest/1.0", options={"myoption": "2"})
        """)
    c.save({"gtest/conanfile.py": gtest,
            "engine/conanfile.py": engine})
    c.run("create gtest")
    c.run("create gtest -o gtest*:myoption=2")
    c.run("create gtest -o gtest*:myoption=3")
    c.run("create engine")
    assert "gtest/1.0: MYOPTION: 2" in c.out
    c.run("create engine -o gtest*:myoption=3")
    assert "gtest/1.0: MYOPTION: 3" in c.out


def test_requires_components():
    """ this test used to fail with "gtest" not required by components
    It is important to have at least 1 external ``requires`` because with
    no requires at all it doesn't fail.
    https://github.com/conan-io/conan/issues/13187
    """
    c = TestClient(light=True)
    conanfile = textwrap.dedent("""
        from conan import ConanFile

        class MyLib(ConanFile):
            name = "mylib"
            version = "0.1"

            requires = "openssl/1.1"
            test_requires = "gtest/1.0"

            def package_info(self):
                self.cpp_info.components["mylib"].requires = ["openssl::openssl"]
        """)
    c.save({"gtest/conanfile.py": GenConanfile("gtest", "1.0"),
            "openssl/conanfile.py": GenConanfile("openssl", "1.1"),
            "pkg/conanfile.py": conanfile})
    c.run("create gtest")
    c.run("create openssl")
    c.run("create pkg")
    # This NO LONGER FAILS
    c.assert_listed_require({"gtest/1.0": "Cache"}, test=True)


def test_requires_transitive_diamond_components():
    """
    libc -----> libb ----> liba
     |-(test-requires)----/
    https://github.com/conan-io/conan/issues/13892
    """
    c = TestClient(light=True)
    libc = textwrap.dedent("""
        from conan import ConanFile
        class LibC(ConanFile):
            name = "libc"
            version = "0.1"
            requires = "libb/1.0"
            test_requires = "liba/1.0"
            def package_info(self):
                 self.cpp_info.components["comp"].libs = ["libc"]
                 self.cpp_info.components["comp"].requires.append("libb::libb")
        """)
    c.save({"liba/conanfile.py": GenConanfile("liba", "1.0"),
            "libb/conanfile.py": GenConanfile("libb", "1.0").with_requires("liba/1.0"),
            "libc/conanfile.py": libc})
    c.run("create liba")
    c.run("create libb")
    c.run("create libc")
    # This used to crash due to component not defined to liba
    assert "libc/0.1: Created package" in c.out


def test_requires_transitive_diamond_components_order():
    """
    like the above, but in different order
     libc  --(test-requires)---> liba
       |-----> libb -----------/
    libc->liba => test=False, direct=True, is_test=True
    direct_dependencies (components check) => False
    https://github.com/conan-io/conan/issues/17164
    """
    c = TestClient(light=True)
    libc = textwrap.dedent("""
        from conan import ConanFile
        class LibC(ConanFile):
            name = "libc"
            version = "0.1"

            def build_requirements(self):
                self.test_requires("liba/1.0")
            def requirements(self):
                self.requires("libb/1.0")

            def package_info(self):
                 self.cpp_info.components["comp"].libs = ["libc"]
                 self.cpp_info.components["comp"].requires.append("libb::libb")
        """)
    c.save({"liba/conanfile.py": GenConanfile("liba", "1.0"),
            "libb/conanfile.py": GenConanfile("libb", "1.0").with_requires("liba/1.0"),
            "libc/conanfile.py": libc})
    c.run("create liba")
    c.run("create libb")
    c.run("create libc")
    # This used to crash due to component not defined to liba
    assert "libc/0.1: Created package" in c.out


def test_wrong_requirement_test_requires():
    """ https://github.com/conan-io/conan/issues/17312

    app --------> etas --------------> enum
      \\-->hwinfo-/---(test_requires)---/
       \\-----------------------------/

    app->enum => test=False, direct=True, is_test=True
    direct_dependencies (components check) => True
    """
    c = TestClient(light=True)
    app = textwrap.dedent("""
        from conan import ConanFile
        class Pkg(ConanFile):
            name = "app"
            version = "0.1"
            requires = "etas/0.1", "enum/0.1", "hwinfo/0.1"

            def generate(self):
                for r, d in self.dependencies.items():
                    assert not (r.direct and r.is_test)

            def package_info(self):
                self.cpp_info.requires = ["etas::etas", "enum::enum", "hwinfo::hwinfo"]
            """)
    files = {
        "enum/conanfile.py": GenConanfile("enum", "0.1"),
        "etas/conanfile.py": GenConanfile("etas", "0.1").with_requirement("enum/0.1"),
        "hwinfo/conanfile.py": GenConanfile("hwinfo", "0.1").with_requirement("etas/0.1")
                                                            .with_test_requires("enum/0.1"),
        "app/conanfile.py": app,
    }
    c.save(files)
    c.run("create enum")
    c.run("create etas")
    c.run("create hwinfo")
    # The assert in generate() doesnt fail
    c.run("install app")
    # the package_info() doesn't fail
    c.run("create app")


def test_test_requires_options():
    """ the default_options = {} values also propagate to ``test_requires()`` """
    c = TestClient(light=True)
    c.save({"test/conanfile.py": GenConanfile("test", "0.1").with_option("myoption", [1, 2, 3]),
            "consumer/conanfile.py": GenConanfile().with_test_requires("test/0.1")
                                                   .with_default_option("test/*:myoption", "2")})
    c.run("create test -o myoption=2")
    c.assert_listed_binary({"test/0.1": ("a3cb1345b8297bfdffea4ef4bb1b2694c54d1d69", "Build")})

    c.run("install consumer")
    c.assert_listed_require({"test/0.1": "Cache"}, test=True)
    c.assert_listed_binary({"test/0.1": ("a3cb1345b8297bfdffea4ef4bb1b2694c54d1d69", "Cache")},
                           test=True)


def test_invisible_requires_options():
    """ the default_options = {} values also propagate to ``requires(visible=False)`` """
    c = TestClient(light=True)
    c.save({"test/conanfile.py": GenConanfile("test", "0.1").with_option("myoption", [1, 2, 3]),
            "consumer/conanfile.py": GenConanfile().with_requirement("test/0.1", visible=False)
                                                   .with_default_option("test/*:myoption", "2")})
    c.run("create test -o myoption=2")
    c.assert_listed_binary({"test/0.1": ("a3cb1345b8297bfdffea4ef4bb1b2694c54d1d69", "Build")})

    c.run("install consumer")
    c.assert_listed_require({"test/0.1": "Cache"})
    c.assert_listed_binary({"test/0.1": ("a3cb1345b8297bfdffea4ef4bb1b2694c54d1d69", "Cache")})



import os
from conan.api.output import Color
from conan.tools.env import Environment
from conan.tools.env.environment import create_env_script


class ROSEnv:
    """
    Generator to serve as integration for Robot Operating System 2 development workspaces.

    IMPORTANT: This generator should be used together with CMakeDeps and CMakeToolchain generators.
    """

    def __init__(self, conanfile):
        """
        :param conanfile: ``< ConanFile object >`` The current recipe object. Always use ``self``.
        """
        self._conanfile = conanfile
        self.variables = {}
        self._build_script_sh_file = "conanrosenv-build.sh"
        self._build_script_bat_file = "conanrosenv-build.bat"
        self._wrapper_script_sh_file = "conanrosenv.sh"
        self._wrapper_script_bat_file = "conanrosenv.bat"

    def generate(self):
        """
        Creates a ``conanrosenv.sh`` with the environment variables that are needed to build and
        execute ROS packages with Conan dependencies.
        """
        cmake_toolchain_path = os.path.join(self._conanfile.generators_folder,
                                            "conan_toolchain.cmake")
        self.variables["CMAKE_TOOLCHAIN_FILE"] = cmake_toolchain_path
        build_type = self._conanfile.settings.get_safe("build_type")
        if build_type:
            self.variables["CMAKE_BUILD_TYPE"] = build_type

        # Add ROS required variables to VirtualBuildEnv
        rosbuildenv = Environment()
        for k, v in self.variables.items():
            rosbuildenv.define(k, v)
        for build_script_file in [self._build_script_sh_file, self._build_script_bat_file]:
            rosbuildenv.vars(self._conanfile, "build").save_script(build_script_file)

        #TODO: Add powrshell support generating .ps1 files
        self._generate_sh_files()
        self._generate_bat_files()

    def _generate_sh_files(self):
        # Generate conanrosenv.sh script wrapper that calls conanbuild.sh and conanrun.sh
        conanbuild_sh_path = os.path.join(self._conanfile.generators_folder, "conanbuild.sh")
        conanrun_sh_path = os.path.join(self._conanfile.generators_folder, "conanrun.sh")
        rosenv_sh_wrapper_content = [f". \"{conanbuild_sh_path}\"", f". \"{conanrun_sh_path}\""]
        create_env_script(self._conanfile, "\n".join(rosenv_sh_wrapper_content),
                          self._wrapper_script_sh_file, None)

        conanrosenv_path = os.path.join(self._conanfile.generators_folder,
                                        self._wrapper_script_sh_file)
        msg = f"Generated ROSEnv Conan file: {self._wrapper_script_sh_file}\n" + \
              f"Use 'source {conanrosenv_path}' to set the ROSEnv Conan before 'colcon build'"
        self._conanfile.output.info(msg, fg=Color.CYAN)

    def _generate_bat_files(self):
        # Generate conanrosenv.bat script wrapper that calls conanbuild.bat and conanrun.bat
        conanbuild_bat_path = os.path.join(self._conanfile.generators_folder, "conanbuild.bat")
        conanrun_bat_path = os.path.join(self._conanfile.generators_folder, "conanrun.bat")
        rosenv_bat_wrapper_content = [
            "@echo off",
            f"CALL \"{conanbuild_bat_path}\"",
            f"CALL \"{conanrun_bat_path}\""
            ]
        create_env_script(self._conanfile, "\n".join(rosenv_bat_wrapper_content),
                          self._wrapper_script_bat_file, None)

        conanrosenv_path = os.path.join(self._conanfile.generators_folder,
                                        self._wrapper_script_bat_file)
        msg = f"Generated ROSEnv Conan file: {self._wrapper_script_bat_file}\n" + \
              f"Use 'call {conanrosenv_path}' to set the ROSEnv Conan before 'colcon build'"
        self._conanfile.output.info(msg, fg=Color.CYAN)



import platform
import textwrap

from conan.test.utils.tools import TestClient


# TODO: Change this test when we refactor EnvVars. The UX leaves much to be desired
def test_env_and_scope_none():
    """
    Check scope=None does not append foo=var to conan{build|run}.{bat|sh|ps1}

    Issue: https://github.com/conan-io/conan/issues/17249
    """
    client = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.env import Environment
        class Pkg(ConanFile):
            name = "pkg"
            version = "0.1"
            settings = "os", "compiler", "build_type", "arch"
            def generate(self):
                env1 = Environment()
                env1.define("foo", "var")
                # Will not append "my_env_file" to "conanbuild.bat|sh|ps1"
                envvars = env1.vars(self, scope=None)
                envvars.save_script("my_env_file")
                # Let's check the apply() function
                with env1.vars(self, scope=None).apply():
                    import os
                    assert os.environ["foo"] == "var"
            """)
    client.save({"conanfile.py": conanfile})
    client.run("install .")
    ext = ".bat" if platform.system() == "Windows" else ".sh"
    assert "my_env_file" not in client.load(f"conanbuild{ext}")
    assert "my_env_file" not in client.load(f"conanrun{ext}")




import json
import textwrap

from conan.test.utils.tools import TestClient


def test_print_in_conanfile():
    """
    Tests that prints in conanfiles will not ruin json stdout outputs
    """
    c = TestClient(light=True)
    other = textwrap.dedent("""
        def myprint(text):
            print(text)
        """)
    conanfile = textwrap.dedent("""
        import other
        from conan import ConanFile

        class MyTest(ConanFile):
            name = "pkg"
            version = "0.1"

            def generate(self):
                print("Hello world!!")
                other.myprint("Bye world!!")
        """)
    c.save({"other.py": other,
            "conanfile.py": conanfile})
    c.run("install . --format=json")
    assert "Hello world!!" in c.stderr
    assert "Bye world!!" in c.stderr
    info = json.loads(c.stdout)
    # the json is correctly loaded
    assert "graph" in info



import pytest

from conan.internal.default_settings import default_settings_yml
from conan.tools.cmake import CMake
from conan.tools.cmake.presets import write_cmake_presets
from conan.internal.model.conf import Conf
from conan.internal.model.settings import Settings
from conan.test.utils.mocks import ConanFileMock, RedirectedTestOutput
from conan.test.utils.test_files import temp_folder
from conan.test.utils.tools import redirect_output


def test_run_install_component():
    """
    Testing that the proper component is installed.
    Issue related: https://github.com/conan-io/conan/issues/10359
    """
    # Load some generic windows settings
    settings = Settings.loads(default_settings_yml)
    settings.os = "Windows"
    settings.arch = "x86"
    settings.build_type = "Release"
    settings.compiler = "msvc"
    settings.compiler.runtime = "dynamic"
    settings.compiler.version = "190"

    conanfile = ConanFileMock()
    conanfile.conf = Conf()
    conanfile.folders.generators = "."
    conanfile.folders.set_base_generators(temp_folder())
    conanfile.settings = settings
    conanfile.folders.set_base_package(temp_folder())

    # Choose generator to match generic settings
    write_cmake_presets(conanfile, "toolchain", "Visual Studio 14 2015", {})
    cmake = CMake(conanfile)
    cmake.install(component="foo")

    assert "--component foo" in conanfile.command


@pytest.mark.parametrize("config, deprecated",
                         [("tools.cmake:install_strip", True),
                          ("tools.build:install_strip", False),])
def test_run_install_strip(config, deprecated):
    """
    Testing that the install/strip rule is called
    Issue related: https://github.com/conan-io/conan/issues/14166
    """

    settings = Settings.loads(default_settings_yml)
    settings.os = "Linux"
    settings.arch = "x86_64"
    settings.build_type = "Release"
    settings.compiler = "gcc"
    settings.compiler.version = "11"

    conanfile = ConanFileMock()

    conanfile.conf = Conf()
    conanfile.conf.define(config, True)

    conanfile.folders.generators = "."
    conanfile.folders.set_base_generators(temp_folder())
    conanfile.settings = settings
    conanfile.folders.set_base_package(temp_folder())

    write_cmake_presets(conanfile, "toolchain", "Unix Makefiles", {})
    cmake = CMake(conanfile)
    stdout = RedirectedTestOutput()  # Initialize each command
    stderr = RedirectedTestOutput()
    with redirect_output(stderr, stdout):
        cmake.install(stdout=stdout, stderr=stderr)

    if deprecated:
        assert "WARN: deprecated: The 'tools.cmake:install_strip' configuration is deprecated, use"\
               " 'tools.build:install_strip' instead" in stderr
    else:
        assert "tools.cmake:install_strip" not in stderr
    assert "--strip" in conanfile.command


def test_run_install_cli_args():
    """
    Testing that the passing cli_args to install works
    Issue related: https://github.com/conan-io/conan/issues/14235
    """

    settings = Settings.loads(default_settings_yml)
    settings.os = "Linux"
    settings.arch = "x86_64"
    settings.build_type = "Release"
    settings.compiler = "gcc"
    settings.compiler.version = "11"

    conanfile = ConanFileMock()

    conanfile.conf = Conf()

    conanfile.folders.generators = "."
    conanfile.folders.set_base_generators(temp_folder())
    conanfile.settings = settings
    conanfile.folders.set_base_package(temp_folder())

    write_cmake_presets(conanfile, "toolchain", "Unix Makefiles", {})
    cmake = CMake(conanfile)
    cmake.install(cli_args=["--prefix=/tmp"])
    assert "--prefix=/tmp" in conanfile.command


def test_run_install_cli_args_strip():
    """
    Testing that the install/strip rule is called when using cli_args
    Issue related: https://github.com/conan-io/conan/issues/14235
    """

    settings = Settings.loads(default_settings_yml)
    settings.os = "Linux"
    settings.arch = "x86_64"
    settings.build_type = "Release"
    settings.compiler = "gcc"
    settings.compiler.version = "11"

    conanfile = ConanFileMock()

    conanfile.conf = Conf()

    conanfile.folders.generators = "."
    conanfile.folders.set_base_generators(temp_folder())
    conanfile.settings = settings
    conanfile.folders.set_base_package(temp_folder())

    write_cmake_presets(conanfile, "toolchain", "Unix Makefiles", {})
    cmake = CMake(conanfile)
    cmake.install(cli_args=["--strip"])
    assert "--strip" in conanfile.command



import textwrap

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


def test_compatible_cppstd():
    c = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile

        class Pkg(ConanFile):
            name = "pkg"
            version = "0.1"
            settings = "os", "compiler"

            def package_info(self):
                self.output.info("PackageInfo!: Cppstd version: %s!"
                                 % self.settings.compiler.cppstd)
        """)
    profile = textwrap.dedent("""
        [settings]
        os = Linux
        compiler=gcc
        compiler.version=12
        compiler.libcxx=libstdc++
        """)
    c.save({"conanfile.py": conanfile,
            "myprofile": profile})
    # Create package with cppstd 17
    c.run("create .  -pr=myprofile -s compiler.cppstd=17")
    package_id = "95dcfeb51c04968b4ee960ee393cf2c1ebcf7782"
    assert f"pkg/0.1: Package '{package_id}' created" in c.out

    # package can be used with a profile gcc cppstd20 falling back to 17
    c.save({"conanfile.py": GenConanfile().with_require("pkg/0.1")})
    c.run("install . -pr=myprofile -s compiler.cppstd=20")
    assert f"Using compatible package '{package_id}'"
    assert "pkg/0.1: PackageInfo!: Cppstd version: 17!" in c.out
    c.assert_listed_binary({"pkg/0.1": (f"{package_id}", "Cache")})
    assert "pkg/0.1: Already installed!" in c.out


def test_compatible_cppstd_missing_compiler():
    c = TestClient()
    settings_user = textwrap.dedent("""
    compiler:
        mycompiler:
            version: ["12"]
            libcxx: ["libstdc++"]
            cppstd: [15, 17, 20]
    """)
    conanfile = textwrap.dedent("""
            from conan import ConanFile

            class Pkg(ConanFile):
                name = "pkg"
                version = "0.1"
                settings = "os", "compiler"

                def package_info(self):
                    self.output.info("PackageInfo!: Cppstd version: %s!"
                                     % self.settings.compiler.cppstd)
            """)
    profile = textwrap.dedent("""
            [settings]
            os = Linux
            compiler=mycompiler
            compiler.version=12
            compiler.libcxx=libstdc++
            """)
    c.save({"conanfile.py": conanfile,
            "myprofile": profile})
    c.save_home({"settings_user.yml": settings_user})
    # Create package with cppstd 17
    c.run("create .  -pr=myprofile -s compiler.cppstd=17")
    package_id = "51a90090adb1cbd330a64b4c3b3b32af809af4f9"
    assert f"pkg/0.1: Package '{package_id}' created" in c.out

    # package can't be used with cppstd 20 and fallback to 17, because mycompiler is not known
    # to the default cppstd_compat function. Ensure that it does not break and warns the user
    c.save({"conanfile.py": GenConanfile().with_require("pkg/0.1")})
    c.run("install . -pr=myprofile -s compiler.cppstd=20", assert_error=True)
    assert "Missing binary: pkg/0.1:b4b07859713551e8aac612f8080888c58b4711ae" in c.out
    assert 'pkg/0.1: WARN: No cppstd compatibility defined for compiler "mycompiler"' in c.out



from conan.errors import ConanInvalidConfiguration, ConanException
from conan.internal.model.version import Version


def check_min_compiler_version(conanfile, compiler_restrictions):
    """(Experimental) Checks if the current compiler and its version meet the minimum requirements.

    :param conanfile: The current recipe object. Always use ``self``.
    :param compiler_restrictions:
        A list of tuples, where each tuple contains:

        - **compiler** (*str*): The name of the compiler (e.g., "gcc", "msvc").
        - **min_version** (*str*): The minimum required version as a string (e.g., "14", "19.0").
        - **reason** (*str*): A string explaining the reason for the version requirement.
    :raises ConanException:
        If the 'compiler' or 'compiler.version' settings are not defined.
    :raises ConanInvalidConfiguration:
        If the found compiler version is less than the specified minimum version for that compiler.

    :Example:
        .. code-block:: python

            def validate(self):
                compiler_restrictions = [
                    ("clang", "14", "requires C++20 coroutines support"),
                    ("gcc", "12", "requires C++20 modules support")
                ]
                check_min_compiler_version(self, compiler_restrictions)
    """
    compiler_value = conanfile.settings.get_safe("compiler")
    if not compiler_value:
        raise ConanException("Called check_min_compiler_version with no compiler defined")
    compiler_version = conanfile.settings.get_safe("compiler.version")
    if not compiler_version:
        raise ConanException("Called check_min_compiler_version with no compiler.version defined")

    for compiler, min_version, reason in compiler_restrictions:
        if compiler_value == compiler:
            if Version(compiler_version) < Version(min_version):
                ref = conanfile.ref if hasattr(conanfile, "ref") else conanfile.name
                raise ConanInvalidConfiguration(
                    f"{ref} requires {compiler} >= {min_version}, but {compiler} {compiler_version} was found\n"
                    f"Reason: {reason}")
            break
